#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

__all__ = ['sn','cz','hi','ef','aed']

# Overview of the metrics (Table 1 in Zheng F et al., [Draft])
#
#   -------------------------------------------------------------------------------------------------------------------
#   Metric       Symbol      Metric range              Run-time search behavior   Comments
#   Category                 Value interpretation      characteristic
#   -------------------------------------------------------------------------------------------------------------------
#   Solution     SN          [0,Infty]                 MOA's ability to find      Wang et al. (2015), but 
#   quality                  larger value is better    solutions that are on      current study is the first to 
#   metric                                             Pareto front               consider its run-time statistics
#   -------------------------------------------------------------------------------------------------------------------
#   Solution     CZ          [0,Infty]                 Relative performance of    Developed by Zheng et al. [Draft]
#   quality                  larger value is better    MOAs in providing non-     
#   metric                                             dominated solutions
#   -------------------------------------------------------------------------------------------------------------------
#   Spacing      EF          [0,1]                     Front extend within the    Developed by Zheng et al. [Draft]
#   metric                   larger value is better    objective space     
#                                                     
#   -------------------------------------------------------------------------------------------------------------------
#   Convergence  HI          [0,1]                     Solution quality,          Hadka and Reed (2012), but
#   metric                   larger value is better    diversity and uniformity   current study is the first to
#                                                                                 consider its run-time variation
#   -------------------------------------------------------------------------------------------------------------------
#   Convergence  AED         [0,1]                     Convergence in objective   Kollat and Reed (2006)
#   metric                   smaller value is better   space   
#                                                      
#   -------------------------------------------------------------------------------------------------------------------
#   Convergence  Var         [0,1]                     Convergence in decision    Introduced in Zharie (2002), but
#   metric                                             space                      current study is the first to
#                                                                                 consider it in multi-objective space
#   -------------------------------------------------------------------------------------------------------------------
#
#
# Literature
# ----------
#        Hadka, D., & Reed, P. M. (2013).
#        Borg: An Auto-Adaptive Many-Objective Evolutionary Computing Framework.
#        Evolutionary Computation, 21(2), 231-259.
#
#        Kollat, J. B., & Reed, P. M. (2006).
#        Comparing state-of-the-art evolutionary multi-objective algorithms for long-term groundwater monitoring design.
#        Advances in Water Resources, 29(6), 792-807. doi:10.1016/j.advwatres.2005.07.010
#
#        Wang, Q., Guidolin, M., Savic, D., & Kapelan, Z. (2015).
#        Two-Objective Design of Benchmark Problems of a Water Distribution System via MOEAs: Towards the Best-Known Approximation of the True Pareto Front.
#        Journal of Water Resources Planning and Management, 141(3), 04014060. doi:10.1061/(ASCE)WR.1943-5452.0000460
#
#        Zaharie, D. (2002).
#        Critical values for the control parameters of differential evolution algorithms (pp. 62-67).
#        Presented at the 8th International Mendel Conference on Soft Computation.
#
#        F Zheng, AC Zecchin, HR Maier, and AR Simpson
#        Comparison of the searching behavior of NSGA-II, SAMODE and Borg MOEAs applied to water
#        distribution system design problems
#        [Draft]

def is_dominated(nobj, point, front):
    # checks whether a new canidate 'point' is dominated by the 'front' or not
    # does not manipulate 'front'
    #
    # returns: 1 --> point is dominating other points on front and hence should be added to the 'front' and others removed
    #          0 --> point is neither dominating others nor dominated by others, ie. it is either part of the 'front' or
    #                is a new point filling gaps or extends front and should hence be added
    #         -1 --> point is dominated
    
    nsets = np.shape(front)[0]

    dominanceFlag = 0
    ii = -1
    while (ii < nsets-1):
        ii += 1
        numeql = 0
        numimp = 0
        numdeg = 0

        for jj in range(nobj):
            if (point[jj] == front[ii,jj]):
                numeql += 1
            if (point[jj] <  front[ii,jj]):
                numimp += 1
            if (point[jj] >  front[ii,jj]):
                numdeg += 1
                
        if (numimp == 0) and (numdeg > 0):
            # 'point' is dominated
            dominanceFlag = -1
            return dominanceFlag
        else:
            if (numeql == nobj):
                # Objective functions are the same for 'point' and archived solution ii
                dominanceFlag = 0
                return dominanceFlag
            else:
                if (numimp > 0) and (numdeg == 0):
                    # 'point' dominates ii-th solution in the front
                    dominanceFlag = 1
                    return dominanceFlag
    
    return dominanceFlag

def point_to_front(nobj, point, front):
    # It checks if 'point' is dominating some points of 'front'.
    # If 'point' dominates others, it adds 'point' to 'front' and deletes dominated ones. At the end the updated 'front'
    # is returned. It hence is doing the same as the routine 'is_dominated', but returns updated front instead of only
    # information about the domination status.
    #
    # (originally routine NDcheck in Matlab version by Masoud)
    #
    # returns: set which only consists of non-dominated points

    dominanceFlag = 0
    
    if len(front) == 0:
        new_set = np.array(point)
        return new_set

    if (len(np.shape(front)) == 1):
        front = np.expand_dims(front, axis=0)

    nsets = np.shape(front)[0]

    ii = -1 
    while (ii < nsets-1):
        ii += 1
        numeql = 0
        numimp = 0
        numdeg = 0

        for jj in range(nobj):
            if (point[jj] == front[ii,jj]):
                numeql += 1
            if (point[jj] <  front[ii,jj]):
                numimp += 1
            if (point[jj] >  front[ii,jj]):
                numdeg += 1
                
        if (numimp == 0) and (numdeg > 0):
            # 'point' is dominated --> return 'front' unchanged
            dominanceFlag = -1
            return front
        else:
            if (numeql == nobj):
                # Objective functions are the same for 'point' and archived solution ii
                # Replace solution ii in 'front' with 'point'
                front[ii] = point
                dominanceFlag = 0
                return front
            else:
                if (numimp > 0) and (numdeg == 0):
                    # 'point' dominates ii-th solution in the 'front'
                    if (nsets > 1):
                        # Remove solution ii from 'front'
                        front = np.delete(front,ii,axis=0)
                        nsets  -= 1
                    else:
                        # Remove last point in 'front'
                        # This means all points on 'front' are dominated by new candidate 'point'
                        front = np.array([])
                        nsets -= 1
                    ii -= 1
                    dominanceFlag = 1

    if (dominanceFlag == 0):
        # 'point' is a new point of the 'front' but does not dominate any current member
        # hence 'point' will be added
        np.vstack([front,np.array(point)])

    if (nsets == 0):
        new_set = np.array(point)
    else:
        new_set = np.vstack((front,np.array(point)))

    return new_set

def sn(front, reference_front):
    """
        (Solution quality metric)
        
        The run-time search quality of a multi-objective optimization algorithm can be measured by the number of solutions
        that form an approxiamte Pareto front at searching time t that are nondominated by the Pareto front Z* in objective space.
        This number of points is called success number SN(t) of the algorithm. SN(t) is basically counting the number of points
        of Z* detected by the algorithm after t function/model evaluations.

        For large problems the reference Pareto front Z* might be hard to determine and is hence often unknown.
        In this case SN can not be determined.

        Comparing two algorithms a larger value of SN(t) indicates the algorithm's greater ability to find solutions along
        the Pareto front Z*.
        

        Definition
        ----------
        def sn(front, reference_front):


        Input
        -----
        front               members of current front
                            array with n1 rows (number of points) and m cols (number of objectives)
        reference_front     all members of the best Pareto front (often unknown)
                            array with n2 rows (number of points) and m cols (number of objectives)


        Optional Input
        --------------
        None


        Output
        ------
        Scalar integer giving the number of points of the current front which are nondominated by the
        reference front. Scalar is less equal n2.


        Literature
        ----------
        F Zheng, AC Zecchin, HR Maier, and AR Simpson
        Comparison of the searching behavior of NSGA-II, SAMODE and Borg MOEAs applied to water
        distribution system design problems
        [Draft]


        Restrictions
        ------------
        Objectives are assumed to be minimized.


        Examples
        --------
        >>> from autostring import astr
        >>> reference_front = np.array([[3,13],[3,11],[3,9],[4,8],[5,7],[6,6],[7,5],[8,4],[10,3],[12,2],[14,1],[16,1],[18,1],[20,1],[22,1]])
        >>> front = np.array([[6,11],[6,8],[7,6],[8,5],[10,3],[12,2],[15,2],[18,2]])
        >>> success_number = sn(front, reference_front)
        >>> print(astr(success_number))
        2
        

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT License.

        Copyright (c) 2016 Juliane Mai

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  JM, Jan 2016 
    """

    # checks if elements of current front are members of reference front
    return np.sum(np.array([ np.where(np.all(reference_front==ff,axis=1))[0].shape[0] for ff in list(front) ]))

def cz(front, all_fronts, all_cz=False):
    """
        (Solution quality metric)
        
        To charcterize and compare the performance of different Multi-objective algorithms (MOAs), the contribution of
        each algorithm to the aggregated best front from all MOEAs can be used.
        
        CZ = Contribution to aggregated best front Z
        

        Definition
        ----------
        def cz(front, all_fronts, all_cz=False):


        Input
        -----
        front               members of current front
                            array with n1 rows (number of points) and m cols (number of objectives)
        all_fronts          fronts of all fronts under consideration
                            these will be aggregated and the amount of points from 'front' in the aggreagtion counted
                            list of k arrays with n2(k) rows (number of points) and m cols (number of objectives)


        Optional Input
        --------------
        all_cz              if True the CZ of every member of all_fronts is returned as a list of integers
                            'front' is then ignored


        Output
        ------
        Scalar integer giving the number of points of the current front which are nondominated by any other point of
        all given fronts 'all_fronts'. Scalar is less equal sum( n2(k) ).
        If 'all_cz' True, then output is list of integers.


        Literature
        ----------
        F Zheng, AC Zecchin, HR Maier, and AR Simpson
        Comparison of the searching behavior of NSGA-II, SAMODE and Borg MOEAs applied to water
        distribution system design problems
        [Draft]


        Restrictions
        ------------
        Objectives are assumed to be minimized.


        Examples
        --------
        >>> from autostring import astr
        >>> front_1 = np.array([[6,11],[6,8],[7,6],[8,5],[10,3],[12,2],[15,2],[18,2]])
        >>> front_2 = np.array([[5,10],[5,9],[6,8],[7,7],[8,6],[9,5],[10,4],[12,3],[14,3],[15,2],[16,1],[18,1]])
        >>> front_3 = np.array([[7,10],[7,8],[8,7],[10,6],[12,5],[13,4],[16,4],[18,3],[20,3]])
        >>> all_fronts = [ front_1, front_2, front_3 ]
        >>> contrib_z = cz(front_1, [ front_1, front_2, front_3 ])
        >>> print(astr(contrib_z))
        5
        >>> contrib_z = cz(front_2, all_fronts)
        >>> print(astr(contrib_z))
        3
        >>> contrib_z = cz(front_3, all_fronts)
        >>> print(astr(contrib_z))
        0
        >>> contrib_z = cz(front_3, all_fronts, all_cz=True)
        >>> print(astr(contrib_z))
        ['5' '3' '0']
        

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT License.

        Copyright (c) 2016 Juliane Mai

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  JM, Feb 2016 
    """
    # aggregate the fronts
    nfronts = len(all_fronts)
    nobj    = all_fronts[0].shape[1]
    aggretated_front = []
    for ifronts in range(nfronts):
        nsets = all_fronts[ifronts].shape[0]
        for isets in range(nsets):
            aggretated_front = point_to_front(nobj, all_fronts[ifronts][isets], aggretated_front) 
        
    # counts elements of current front which are members of aggregated front
    if not(all_cz):
        return np.sum(np.array([ np.where(np.all(aggretated_front==ff,axis=1))[0].shape[0] for ff in list(front) ]))
    else:
        return [ np.sum(np.array([ np.where(np.all(aggretated_front==ff,axis=1))[0].shape[0] for ff in list(kk) ])) for kk in all_fronts ]


def hi(front, reference_point, nsamples=None, reference_front=None, hi_range=False):
    """
        (Convergence metric)
        
        The hypervolume indicator (HI) is a measure for soultion distribution and uniformity in objective space (Zitzler and Thiele, 1999).
        It measures the hypervolume of the multi-dimensional region enclosed by a front with respect to a reference point.
        It represents the overall searching performance of solution quality, solution diversity and the uniformity of the solutions on the
        front (Hadka and Reed, 2012).

        By default, the routine returns an estimation of the hypervolume dominated by the approximated Pareto front
        set 'front' and bounded by the reference point 'reference_point'. This means the volume covered by the 'front' relative to the
        overall feasible hypervolume.

                ^
                |
             10 |---------------------------------------* (reference point)
                |                 |                     |
                |                 |       A ~ 20        |
                |                 *                     |              ==> HI = 20 / 50 = 0.4
                |                  *                    |
                |            (front) *                  |
                |                        *--------------|
                |                                       |
                -------------------------------------------------->
                                                        5

        Optionally, HI can be returned as the ratio of the hypervolume relative to that of the known best Pareto front Z* when
        argument 'reference front' is given. Hence, HI lies within [0,1] which a large value representing a hypervolume closer to Z*.

                ^
                |
             10 |---------------------------------------* (reference point)
                |       |         |                     |
                |       | B ~ 30  |      A ~ 20         |
                |       *         *                     |              ==> HI = A / B = 20 / 30 = 0.667
                |        *         *   (front)          |
                |         *          *                  |
                |   (best)  *            *--------------|
                |               *-----------------------|
                -------------------------------------------------->
                                                        5        

        Definition
        ----------
        def hi(front, reference_point, nsamples=None, reference_front=None, hi_range=False):


        Input
        -----
        front               members of current front
                            array with n1 rows (number of points) and m cols (number of objectives)
        reference_point     worst possible point in objective space


        Optional Input
        --------------
        nsamples            number of random points used to approximate the hypervolume
                            (default: 1000)
        reference_front     best known pareto front
                            if given the HI is returned as the ratio between HI of tested front and HI of the reference front
                            value is hence bounded between 0 and 1 where large values indicate a front closer to the reference front
        hi_range            if True an lower and upper value for HI will be returned
                            this range is due to pointwise given 'front', i.e. the coarser the 'front' the more randomly sampled points
                            have a dominance flag of zero and the more uncertain is the HI

                            
        Output
        ------
        if( reference_front == None):
            Scalar floating number of the ratio between hypervolume covered by the 'front' and
            hypervolume bounded by 'reference point'
        else:
            Scalar floating number of the ratio between hypervolume covered by the 'front' and
            hypervolume covered by 'reference front'

            
        Literature
        ----------
        F Zheng, AC Zecchin, HR Maier, and AR Simpson
        Comparison of the searching behavior of NSGA-II, SAMODE and Borg MOEAs applied to water distribution system design problems
        [Draft]

        Hadka, D., and Reed, P. (2012).
        'Diagnostic Assessment of Search Controls and Failure Modes in Many-Objective Evolutionary Optimization.'
        Evolutionary Computation, 20(3), 423-452.

        Zitzler, E., and Thiele, L. (1999).
        'Multiobjective evolutionary algorithms: a comparative case study and the strength Pareto approach.'
        Evolutionary Computation, IEEE Transactions on, 3(4), 257-271.

        
        Restrictions
        ------------
        Objectives are assumed to be minimized.


        Examples
        --------
        >>> from autostring import astr
        >>> front           = np.array([[10,12],[12.5,11],[15,10],[17.5,9],[20,8]])
        >>> reference_point = np.array([25,15])
        >>> reference_front = np.array([[8,12],[11,10],[14,8],[17,6],[20,4]])
        >>> hypervolume_indicator = hi(front, reference_point, nsamples=10000)
        >>> # theoretical value: 85/375 = 0.226667
        >>> print(astr(hypervolume_indicator,prec=4))
        0.2321
        >>> hypervolume_indicator = hi(front, reference_point, nsamples=10000, hi_range=True)
        >>> # theoretical value: 85/375 = 0.226667
        >>> print(astr(hypervolume_indicator,prec=4))
        ['0.2195' '0.2448']
        >>> hypervolume_indicator = hi(front, reference_point, reference_front=reference_front, nsamples=10000)
        >>> # theoretical value: 85/139 = 0.6115
        >>> print(astr(hypervolume_indicator,prec=4))
        0.6258
        >>> hypervolume_indicator = hi(front, reference_point, reference_front=reference_front, nsamples=10000, hi_range=True)
        >>> # theoretical value: 85/139 = 0.6115
        >>> print(astr(hypervolume_indicator,prec=4))
        ['0.5440' '0.7234']
        >>> front           = np.array([[6,11],[6,8],[7,6],[8,5],[10,3],[12,2],[15,2],[18,2]])
        >>> reference_point = np.array([25,15])
        >>> reference_front = np.array([[3,13],[3,11],[3,9],[4,8],[5,7],[6,6],[7,5],[8,4],[10,3],[12,2],[14,1],[16,1],[18,1],[20,1],[22,1]])
        >>> hypervolume_indicator = hi(front, reference_point)
        >>> print(astr(hypervolume_indicator,prec=4))
        0.6180
        >>> hypervolume_indicator = hi(front, reference_point, reference_front=reference_front)
        >>> print(astr(hypervolume_indicator,prec=4))
        0.8477
        

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT License.

        Copyright (c) 2016 Juliane Mai

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  JM, Feb 2016 
    """

    import random
    from autostring import astr

    # check for optionals
    if (nsamples is None):
        nsamp = 1000
    else:
        nsamp = nsamples
    

    # initialization
    nsets_front = front.shape[0]
    nobj        = front.shape[1]
    random.seed(1000)

    # add edge points to 'front'
    best_all_directions = np.min(front,axis=0)
    for iobj in range(nobj):
        edge = best_all_directions.copy()
        edge[iobj] = reference_point[iobj]
        # print(edge)
        # Or is it ??
        # edge = reference_point
        # edge[iobj] = best_all_directions[iobj]
        front = np.vstack([front,np.array(edge)])

    # add edge points to 'reference_front'
    if ( not(reference_front is None) ):
        best_all_directions = np.min(reference_front,axis=0)
        for iobj in range(nobj):
            edge = best_all_directions.copy()
            edge[iobj] = reference_point[iobj]
            # print(edge)
            # Or is it ??
            # edge = reference_point
            # edge[iobj] = best_all_directions[iobj]
            reference_front = np.vstack([reference_front,np.array(edge)])

    # draw random points in feasible domain bounded by 0 and 'reference_point'
    mc_sample = [ [ random.random()*reference_point[ii] for ii in range(nobj) ] for jj in range(nsamp) ]

    # hypervolume of 'front'
    dom_flags = [ is_dominated(nobj, sample, front) for sample in mc_sample ]
    n_below_front = np.where(np.array(dom_flags)== 1)[0].shape[0]
    n_above_front = np.where(np.array(dom_flags)==-1)[0].shape[0]   # this will be the hypervolume
    n_on_front    = np.where(np.array(dom_flags)== 0)[0].shape[0]   # should never happen, otherwise adding of edges is wrong or it is an integer problem

    # front_file = open('hypervolume_front.out','w')
    # for isamp in range(nsamp):
    #     front_file.write(astr(dom_flags[isamp]) + '  ' + astr(mc_sample[isamp][0],prec=4) + '  ' + astr(mc_sample[isamp][1],prec=4) + '  ' +  '\n') 
    # front_file.close()

    # hypervolume of 'reference_front'
    if ( not(reference_front is None) ):
        dom_flags = [ is_dominated(nobj, sample, reference_front) for sample in mc_sample ]
        n_below_reffront = np.where(np.array(dom_flags)== 1)[0].shape[0]
        n_above_reffront = np.where(np.array(dom_flags)==-1)[0].shape[0]   # this will be the hypervolume
        n_on_reffront    = np.where(np.array(dom_flags)== 0)[0].shape[0]   # should never happen, otherwise adding of edges is wrong or it is an integer problem

        # reffront_file = open('hypervolume_reffront.out','w')
        # for isamp in range(nsamp):
        #     reffront_file.write(astr(dom_flags[isamp]) + '  ' + astr(mc_sample[isamp][0],prec=4) + '  ' + astr(mc_sample[isamp][1],prec=4) + '  ' + '\n') 
        # reffront_file.close()
    
    if (reference_front is None):
        if hi_range:
            hi = np.array([(1.0*n_above_front + 0.0*n_on_front) / (1.0*nsamp),
                           (1.0*n_above_front + 1.0*n_on_front) / (1.0*nsamp) ])
        else:
            hi = (1.0*n_above_front + 0.5*n_on_front) / (1.0*nsamp)
    else:
        if hi_range:
            hi = np.array([(1.0*n_above_front + 0.0*n_on_front) / (1.0*n_above_reffront + 1.0*n_on_reffront),
                           (1.0*n_above_front + 1.0*n_on_front) / (1.0*n_above_reffront + 0.0*n_on_reffront) ])
        else:
            hi = (1.0*n_above_front + 0.5*n_on_front) / (1.0*n_above_reffront + 0.5*n_on_reffront)
        
    return hi


def ef(front, reference_front):
    """
        (Spacing metric)
        
        This metric measures the approximate extend of the front (EF) relative to the known best Pareto front Z* given by

             EF = max_(i,j = 1,...,N) ( dist(F_i, F_j) ) / max_(F,G in Z*) ( dist(F, G) )

        where the nominator indicates the maximum distance between two points on an algorithm's front and the denominator
        is the maximum distance between two points on the known Pareto front Z*. dist(X,Y) is the Euclidean distance between
        two vectors X and Y. To calculate EF, each dimension of the solution vectors in Z* is normalized to [0,1] initially,
        followed by the normalization on each dimension of the 'front' Z using the data ranges from Z*.

        A larger value of EF suggests that the approximate front is closer to the Pareto front in terms of solution extend.
        It should be noted that the proposed EF only measures the coverage of the two-objective Pareto approximate fronts, and
        that the overall quality of the fronts has to be assessed using other metrics.
   

        Definition
        ----------
        def ef(front, reference_front)


        Input
        -----
        front               members of current front
                            array with n1 rows (number of points) and m cols (number of objectives)
        reference_front     all members of the best Pareto front (often unknown)
                            array with n2 rows (number of points) and m cols (number of objectives)


        Optional Input
        --------------
        None

                            
        Output
        ------
        Scalar floating number determining the extend of the 'front' relative to the extend of the 'reference_front'.

            
        Literature
        ----------
        F Zheng, AC Zecchin, HR Maier, and AR Simpson
        Comparison of the searching behavior of NSGA-II, SAMODE and Borg MOEAs applied to water distribution system design problems
        [Draft]

        
        Restrictions
        ------------
        Objectives are assumed to be minimized.


        Examples
        --------
        >>> from autostring import astr
        >>> reference_front = np.array([[3,13],[3,11],[3,9],[4,8],[5,7],[6,6],[7,5],[8,4],[10,3],[12,2],[14,1],[16,1],[18,1],[20,1],[22,1]])
        >>> front_1 = np.array([[6,11],[6,8],[7,6],[8,5],[10,3],[12,2],[15,2],[18,2]])
        >>> front_2 = np.array([[5,10],[5,9],[6,8],[7,7],[8,6],[9,5],[10,4],[12,3],[14,3],[15,2],[16,1],[18,1]])
        >>> front_3 = np.array([[7,10],[7,8],[8,7],[10,6],[12,5],[13,4],[16,4],[18,3],[20,3]])
        >>> extend_front = ef(front_1, reference_front)
        >>> print(astr(extend_front,prec=4))
        0.6933
        >>> extend_front = ef(front_2, reference_front)
        >>> print(astr(extend_front,prec=4))
        0.7179
        >>> extend_front = ef(front_3, reference_front)
        >>> print(astr(extend_front,prec=4))
        0.6358
        

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT License.

        Copyright (c) 2016 Juliane Mai

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  JM, Feb 2016 
    """
    from scipy.spatial.distance import pdist

    nobj = reference_front.shape[1]
    if ( nobj != 2 ):
        raise ValueError('pareto_metrics: This metric is only applicable for 2-dimensional fronts!')

    # assure that input is floating and not integer
    reference_front = reference_front.astype(float)
    front           = front.astype(float)

    # column-wise range of reference-front used to scale 'front' and 'reference_front'
    col_min = np.min(reference_front,axis=0)
    col_max = np.max(reference_front,axis=0)
    
    reference_front     = (reference_front - col_min ) / (col_max - col_min)
    front               = (          front - col_min ) / (col_max - col_min)
    extend_front = np.max(pdist(front)) / np.max(pdist(reference_front))
    return extend_front


def aed(front, reference_front):
    """
        (Convergence metric)
        
        The average Euclidean distance (AED) between an approximate front and the known best Pareto front Z* is, besides HI,
        typically used to represent an MOA's convergence status in the objective space (Kollat and Reed, 2006).
        The average Euclidean distance represents the average of all minimal Euclidean distances between a point of the 
        'front' to all points of the 'reference front'.

            AED = 1/N * sum_{i=1}^N min( dist(F_i, F) ),  where dist(F_i,F) is Euclidean distance of point F_i of 'front'
                                                                            to the points of 'reference front' F

        To allow for a comparison of AED values of different 'fronts', a normalization of the AED is applied. Therefore, all
        dimensions of the reference front Z* are normalized using the minimal and maxiaml values of each dimension. All
        dimensions are hence normaized to [0,1]. The same normalization is subsequently applied to the dimensions 'front' using the
        minimal and maximal values of the 'reference front'. The AED is thus within the range [0,1] where a lower value indicates a
        better front in general, as it possesses an overall shorter distance to the Pareto front in objective space.

        It should be noted that this metric does not take the extend of the 'front' into account as it measures the mean of the
        Euclidean distance between each solution F_i of the 'front' and its corresponding nearest member in the refence front Z*.
   

        Definition
        ----------
        def aed(front, reference_front)


        Input
        -----
        front               members of current front
                            array with n1 rows (number of points) and m cols (number of objectives)
        reference_front     all members of the best Pareto front (often unknown)
                            array with n2 rows (number of points) and m cols (number of objectives)


        Optional Input
        --------------
        None

                            
        Output
        ------
        Scalar floating number determining the average euclidean distance of the 'front' to the 'reference_front'

            
        Literature
        ----------
        F Zheng, AC Zecchin, HR Maier, and AR Simpson
        Comparison of the searching behavior of NSGA-II, SAMODE and Borg MOEAs applied to water distribution system design problems
        [Draft]

        
        Restrictions
        ------------
        Objectives are assumed to be minimized.


        Examples
        --------
        >>> from autostring import astr
        >>> reference_front = np.array([[3,13],[3,11],[3,9],[4,8],[5,7],[6,6],[7,5],[8,4],[10,3],[12,2],[14,1],[16,1],[18,1],[20,1],[22,1]])
        >>> front_1 = np.array([[6,11],[6,8],[7,6],[8,5],[10,3],[12,2],[15,2],[18,2]])
        >>> front_2 = np.array([[5,10],[5,9],[6,8],[7,7],[8,6],[9,5],[10,4],[12,3],[14,3],[15,2],[16,1],[18,1]])
        >>> front_3 = np.array([[7,10],[7,8],[8,7],[10,6],[12,5],[13,4],[16,4],[18,3],[20,3]])
        >>> avergae_euclid_dist = aed(front_1, reference_front)
        >>> print(astr(avergae_euclid_dist,prec=4))
        0.0071
        >>> avergae_euclid_dist = aed(front_2, reference_front)
        >>> print(astr(avergae_euclid_dist,prec=4))
        0.0090
        >>> avergae_euclid_dist = aed(front_3, reference_front)
        >>> print(astr(avergae_euclid_dist,prec=4))
        0.0341
        

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT License.

        Copyright (c) 2016 Juliane Mai

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  JM, Feb 2016 
    """

    # assure that input is floating and not integer
    reference_front = reference_front.astype(float)
    front           = front.astype(float)

    # column-wise range of reference-front used to scale 'front' and 'reference_front'
    col_min = np.min(reference_front,axis=0)
    col_max = np.max(reference_front,axis=0)
    
    reference_front     = (reference_front - col_min ) / (col_max - col_min)
    front               = (          front - col_min ) / (col_max - col_min)
    average_euclid_dist = np.mean([ np.min(np.sum((reference_front-ifront)**2,axis=1)) for ifront in front ])
    return average_euclid_dist



if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



    





#!/usr/bin/env python
"""
Convert strings to LaTeX strings in math environment used by matplotlib's
usetex

This module was written by Matthias Cuntz while at Department of Computational
Hydrosystems, Helmholtz Centre for Environmental Research - UFZ, Leipzig,
Germany, and continued while at Institut National de Recherche pour
l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2015-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   str2tex

History
    * Written Oct 2015 by Matthias Cuntz (mc (at) macu (dot) de)
    * Use raw strings for escaped characters, Nov 2021, Matthias Cuntz
    * Bug in space2linebreak in complex strings with spaces;
      do space2linebreak first, Nov 2021, Matthias Cuntz
    * Bug in escaping %, Nov 2021, Matthias Cuntz
    * Remove trailing $\\mathrm{}$, Nov 2021, Matthias Cuntz
    * Ported into pyjams, Nov 2021, Matthias Cuntz
    * Better handling of linebreaks in Matplotlib and LaTeX mode,
      Nov 2021, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz
    * Use input2array, array2input, Jun 2022, Matthias Cuntz
    * Do not escape % if not usetex, Apr 2023, Matthias Cuntz
    * Add unicode symbol degree \u00B0, which get replaced by ^\\circ
      if usetex==True, Apr 2023, Matthias Cuntz

"""
from .helper import input2array, array2input


__all__ = ['str2tex']


def str2tex(strin, space2linebreak=False,
            bold=False, italic=False, usetex=False):
    """
    Convert strings to LaTeX strings in math environment used by matplotlib's
    usetex

    Strings are embedded into '$\\mathrm{strin}$' by default but can be
    embedded into '\\mathbf' and '\\mathit'.
    Spaces are escaped but can be replaced by linebreaks.

    Parameters
    ----------
    strin : str or array-like of str
        string (array)
    space2linebreak : bool, optional
        Replace space (' ') by linebreak ('\\n') if True (default: False)
    bold : bool, optional
        Use '\\mathbf' instead of '\\mathrm' if True (default: False)
    italic : bool, optional
        Use '\\mathit' instead of '\\mathrm' if True (default: False)
    usetex : bool, optional
        Treat only linebreaks and comments if False (default)

    Returns
    -------
    string : str
        string (array) that can be used in matplotlib independent of usetex.

    Examples
    --------
    .. code-block:: python

       fig = plt.figure()
       tit = str2tex('A $S_{Ti}$ is great\\nbut use-less', usetex=usetex)
       fig.suptitle(tit)

    """
    import matplotlib.pyplot as plt

    # Input type and shape
    # use list because numpy array cannot change to raw strings
    istrin = list(input2array(strin, undef='', default=''))

    # font style
    if (bold+italic) > 1:
        raise ValueError('bold and italic are mutually exclusive.')
    else:
        if bold:
            mtex = r'$\mathbf{'
            ttex = r'$\textbf{'
            empty = r'$\mathbf{}$'
        elif italic:
            mtex = r'$\mathit{'
            ttex = r'$\textit{'
            empty = r'$\mathit{}$'
        else:
            mtex = r'$\mathrm{'
            ttex = r'$\textrm{'
            empty = r'$\mathrm{}$'

    # helpers
    a0 = chr(0)  # ascii 0
    # string replacements
    if usetex:
        # no '\n' in LaTeX, use '\newline'
        rep_n       = lambda s: s.replace(r'\n', '}$' + a0 + r'\newline'
                                          + a0 + mtex)
        rep_newline = lambda s: s.replace(r'\newline', '}$' + a0 + r'\newline'
                                          + a0 + mtex)
    else:
        # '\n' has to be unicode string and not raw string in Matplotlib
        rep_n       = lambda s: s.replace(r'\n', '' + a0 + '\n'
                                          + a0 + '')
        rep_newline = lambda s: s.replace(r'\newline', '' + a0 + '\n'
                                          + a0 + '')
    rep_down     = lambda s: s.replace('_', r'\_')
    rep_up       = lambda s: s.replace('^', r'\^')
    rep_hash     = lambda s: s.replace('#', r'\#')
    rep_percent  = lambda s: s.replace('%', r'\%')
    rep_nopercent = lambda s: s.replace(r'\%', '%')
    rep_space    = lambda s: s.replace(' ', r'\ ')
    rep_minus    = lambda s: s.replace('-', '}$' + ttex + '-}$' + mtex)
    rep_degree   = lambda s: s.replace(u'\u00B0', r'^\circ{}')
    rep_a02empty = lambda s: s.replace(a0, '')
    if usetex or (plt.get_backend() == 'pdf'):
        rep_space2n = lambda s: s.replace(' ', '}$' + a0 + r'\newline'
                                          + a0 + mtex)
    else:
        rep_space2n = lambda s: s.replace(' ', ''+'\n'+'')
    rep_empty = lambda s: s.replace(empty, '')

    if usetex:
        for j, s in enumerate(istrin):
            if '$' in s:
                cleanempty = empty not in s
                ss = s.split('$')
                # outside $...$
                # -, _, ^ only escaped if not between $
                for ii in range(0, len(ss), 2):
                    ss[ii] = mtex + ss[ii] + '}$'
                    # - not minus sign
                    if '-' in ss[ii]:
                        ss[ii] = rep_minus(ss[ii])
                        if ss[ii].endswith('{}$'):
                            ss[ii] = ss[ii][:-11]  # rm trailing $\mathrm{}$
                    # \n not in tex mode but normal matplotlib
                    if (r'\n' in ss[ii]) and not (r'\newline' in ss[ii]):
                        ss[ii] = rep_n(ss[ii])
                    elif (r'\newline' in ss[ii]):
                        ss[ii] = rep_newline(ss[ii])
                    # escape _
                    if '_' in ss[ii]:
                        ss[ii] = rep_down(ss[ii])
                    # escape ^
                    if '^' in ss[ii]:
                        ss[ii] = rep_up(ss[ii])
                    # escape #
                    if '#' in ss[ii]:
                        ss[ii] = rep_hash(ss[ii])
                    # escape %
                    if ('%' in ss[ii]) and not (r'\%' in ss[ii]):
                        ss[ii] = rep_percent(ss[ii])
                    # replace unicode degree symbol (after escape of ^)
                    if u'\u00B0' in ss[ii]:
                        ss[ii] = rep_degree(ss[ii])
                    if space2linebreak:
                        if ' ' in ss[ii]:
                            ss[ii] = rep_space2n(ss[ii])

                # reassemble string
                istrin[j] = '$'.join(ss)

                if s[0] == '$':
                    # rm leading $\mathrm{}$ if string started with $
                    istrin[j] = istrin[j][11:]
            else:
                cleanempty = True
                istrin[j] = mtex + s + r'}$'
                # - not minus sign
                if '-' in istrin[j]:
                    istrin[j] = rep_minus(istrin[j])
                    if istrin[j].endswith('{}$'):
                        istrin[j] = istrin[j][:-11]  # rm trailing $\mathrm{}$
                # \n not in tex mode but normal matplotlib
                if (r'\n' in istrin[j]) and not (r'\newline' in istrin[j]):
                    istrin[j] = rep_n(istrin[j])
                elif (r'\newline' in istrin[j]):
                    istrin[j] = rep_newline(istrin[j])
                # escape _
                if '_' in istrin[j]:
                    istrin[j] = rep_down(istrin[j])
                # escape ^
                if '^' in istrin[j]:
                    istrin[j] = rep_up(istrin[j])
                # escape #
                if '#' in istrin[j]:
                    istrin[j] = rep_hash(istrin[j])
                # escape %
                if ('%' in istrin[j]) and not (r'\%' in istrin[j]):
                    istrin[j] = rep_percent(istrin[j])
                # replace unicode degree symbol (after escape of ^)
                if u'\u00B0' in istrin[j]:
                    istrin[j] = rep_degree(istrin[j])
                if space2linebreak:
                    if ' ' in istrin[j]:
                        istrin[j] = rep_space2n(istrin[j])

            # rm $\mathrm{}$
            if cleanempty:
                istrin[j] = rep_empty(istrin[j])

            # escape space
            if ' ' in istrin[j]:
                istrin[j] = rep_space(istrin[j])

            # rm ascii character 0 around linebreaks introduced above
            if a0 in istrin[j]:
                istrin[j] = rep_a02empty(istrin[j])
    else:
        # # escape %
        # istrin = [ rep_percent(i) if ('%' in i) and not (r'\%' in i) else i
        #            for i in istrin ]
        # do not escape %
        istrin = [ rep_nopercent(i) if (r'\%' in i) else i for i in istrin ]
        # '\n' is Matplotlib but nor LaTeX
        for j, s in enumerate(istrin):
            if (r'\n' in istrin[j]) and not (r'\newline' in istrin[j]):
                istrin[j] = rep_n(istrin[j])
            elif (r'\newline' in istrin[j]):
                istrin[j] = rep_newline(istrin[j])
            if a0 in istrin[j]:
                istrin[j] = rep_a02empty(istrin[j])
            if space2linebreak:
                if ' ' in istrin[j]:
                    istrin[j] = rep_space2n(istrin[j])
        if (plt.get_backend() == 'pdf'):  # pragma: no cover
            # pdf backend uses LaTeX
            istrin = [ i.replace(r'\n', r'\newline')
                       if (r'\n' in i) and not (r'\newline' in i) else i
                       for i in istrin ]

    # Return right type
    out = array2input(istrin, strin, undef='')
    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # strin = ['One', 'One-', 'One-Two', 'One Two', 'One\nTwo', 'A $S_{Ti}$ is great\nbut use-less']
    # print(str2tex(strin))
    # # ['$\\mathrm{One}$', '$\\mathrm{One}$$\\textrm{-}$', '$\\mathrm{One}$$\\textrm{-}$$\\mathrm{Two}$', '$\\mathrm{One\\ Two}$', '$\\mathrm{One}$ \n $\\mathrm{Two}$', '$\\mathrm{A\\ }$$S_{Ti}$$\\mathrm{\\ is\\ great}$ \n $\\mathrm{but\\ use}$$\\textrm{-}$$\\mathrm{less}$']
    # print(str2tex(strin, bold=True))
    # # ['$\\mathbf{One}$', '$\\mathbf{One}$$\\textbf{-}$', '$\\mathbf{One}$$\\textbf{-}$$\\mathbf{Two}$', '$\\mathbf{One\\ Two}$', '$\\mathbf{One}$ \n $\\mathbf{Two}$', '$\\mathbf{A\\ }$$S_{Ti}$$\\mathbf{\\ is\\ great}$ \n $\\mathbf{but\\ use}$$\\textbf{-}$$\\mathbf{less}$']
    # print(str2tex(strin, italic=True))
    # # ['$\\mathit{One}$', '$\\mathit{One}$$\\textit{-}$', '$\\mathit{One}$$\\textit{-}$$\\mathit{Two}$', '$\\mathit{One\\ Two}$', '$\\mathit{One}$ \n $\\mathit{Two}$', '$\\mathit{A\\ }$$S_{Ti}$$\\mathit{\\ is\\ great}$ \n $\\mathit{but\\ use}$$\\textit{-}$$\\mathit{less}$']
    # print(str2tex(strin, space2linebreak=True))
    # # ['$\\mathrm{One}$', '$\\mathrm{One}$$\\textrm{-}$', '$\\mathrm{One}$$\\textrm{-}$$\\mathrm{Two}$', '$\\mathrm{One}$ \n $\\mathrm{Two}$', '$\\mathrm{One}$ \n $\\mathrm{Two}$', '$\\mathrm{A}$ \n $\\mathrm{}$$S_{Ti}$$\\mathrm{ \n $\\mathrm{is \n $\\mathrm{great}$ \n $\\mathrm{but \n $\\mathrm{use}$$\\textrm{-}$$\\mathrm{less}$']
    # print(str2tex(strin, space2linebreak=True, bold=True))
    # # ['$\\mathbf{One}$', '$\\mathbf{One}$$\\textbf{-}$', '$\\mathbf{One}$$\\textbf{-}$$\\mathbf{Two}$', '$\\mathbf{One}$ \n $\\mathbf{Two}$', '$\\mathbf{One}$ \n $\\mathbf{Two}$', '$\\mathbf{A}$ \n $\\mathbf{}$$S_{Ti}$$\\mathbf{ \n $\\mathbf{is \n $\\mathbf{great}$ \n $\\mathbf{but \n $\\mathbf{use}$$\\textbf{-}$$\\mathbf{less}$']
    # print(str2tex(strin, usetex=False))
    # # ['One', 'One-', 'One-Two', 'One Two', 'One\nTwo', 'A $S_{Ti}$ is great\nbut use-less']
    # print(str2tex(strin, space2linebreak=True, usetex=False))
    # # ['One', 'One-', 'One-Two', 'One\nTwo', 'One\nTwo', 'A\n$S_{Ti}$\nis\ngreat\nbut\nuse-less']

    # strin = [r'A $S_{Ti}$ is great\nbut use-less-']
    # outsp = [r'A\n$S_{Ti}$\nis\ngreat\nbut\nuse-less-']
    # outsp = [ s.replace(r'\n', '' + '\n' + '') for s in outsp ]
    # print(outsp)
    # print(str2tex(strin, space2linebreak=True))

    # strin = [r'A $S_{Ti}$ is great\nbut use-less-']
    # outsp = [r'$\mathrm{A}$\newline$S_{Ti}$\newline$\mathrm{is}$'
    #          r'\newline$\mathrm{great}$\newline$\mathrm{but}$'
    #          r'\newline$\mathrm{use}$$\textrm{-}$$\mathrm{less}$'
    #          r'$\textrm{-}$']
    # print(outsp)
    # print(str2tex(strin, space2linebreak=True, usetex=True))



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np
"""
    Intersection of two curves from x,y coordinates.

    Inspired from Matlab code
        http://uk.mathworks.com/matlabcentral/fileexchange/11837-fast-and-robust-curve-intersections


    Example
    -------
        a, b = 1, 2
        phi  = np.linspace(3, 10, 100)
        x1   = a*phi - b*np.sin(phi)
        y1   = a - b*np.cos(phi)

        x2   = phi
        y2   = np.sin(phi)+2
        x, y = intersection(x1, y1, x2, y2)

        plt.plot(x1, y1, c="r")
        plt.plot(x2, y2, c="g")
        plt.plot(x, y, "*k")
        plt.show()


    Licence
    -------
    This file is part of the JAMS Python package, distributed under the MIT License.

    Copyright (c) 2017 Sukhbinder Singh

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

    Sukhbinder
    5 April 2017
"""

__all__ = ['intersection']

def _rect_inter_inner(x1, x2):
    n1 = x1.shape[0]-1
    n2 = x2.shape[0]-1
    X1 = np.c_[x1[:-1],x1[1:]]
    X2 = np.c_[x2[:-1],x2[1:]]
    
    S1 = np.tile(X1.min(axis=1),(n2,1)).T
    S2 = np.tile(X2.max(axis=1),(n1,1))
    S3 = np.tile(X1.max(axis=1),(n2,1)).T
    S4 = np.tile(X2.min(axis=1),(n1,1))

    return S1, S2, S3, S4


def _rectangle_intersection_(x1, y1, x2, y2):
    S1, S2, S3, S4 = _rect_inter_inner(x1,x2)
    S5, S6, S7, S8 = _rect_inter_inner(y1,y2)

    C1 = np.less_equal(S1,S2)
    C2 = np.greater_equal(S3,S4)
    C3 = np.less_equal(S5,S6)
    C4 = np.greater_equal(S7,S8)

    ii,jj = np.nonzero(C1 & C2 & C3 & C4)

    return ii, jj


def intersection(x1, y1, x2, y2):
    """
        Intersections of curves given by x,y coordinates.

        Computes the (x,y) locations where two curves intersect.
        The curves can be broken with NaNs or have vertical segments.


        Definition
        ----------
        def intersection(x1, y1, x2, y2):


        Input
        -----
        x1, y1     x,y coordinates of first curve
        x2, y2     x,y coordinates of second curve


        Output
        ------
        x,y = intersection(x1, y1, x2, y2)
        x,y coordinates of all intersection points


        Theory
        ------
        From the Matlab code:
        
        Given two line segments, L1 and L2,

          L1 endpoints:  (x1(1),y1(1)) and (x1(2),y1(2))
          L2 endpoints:  (x2(1),y2(1)) and (x2(2),y2(2))

        we can write four equations with four unknowns and then solve them.  The
        four unknowns are t1, t2, x0 and y0, where (x0,y0) is the intersection of
        L1 and L2, t1 is the distance from the starting point of L1 to the
        intersection relative to the length of L1 and t2 is the distance from the
        starting point of L2 to the intersection relative to the length of L2.

        So, the four equations are

           (x1(2) - x1(1))*t1 = x0 - x1(1)
           (x2(2) - x2(1))*t2 = x0 - x2(1)
           (y1(2) - y1(1))*t1 = y0 - y1(1)
           (y2(2) - y2(1))*t2 = y0 - y2(1)

        Rearranging and writing in matrix form,

         [x1(2)-x1(1)       0       -1   0;      [t1;      [-x1(1);
               0       x2(2)-x2(1)  -1   0;   *   t2;   =   -x2(1);
          y1(2)-y1(1)       0        0  -1;       x0;       -y1(1);
               0       y2(2)-y2(1)   0  -1]       y0]       -y2(1)]

        Let's call that A*T = B.  We can solve for T with T = A\B.

        Once we have our solution we just have to look at t1 and t2 to determine
        whether L1 and L2 intersect.  If 0 <= t1 < 1 and 0 <= t2 < 1 then the two
        line segments cross and we can include (x0,y0) in the output.

        In principle, we have to perform this computation on every pair of line
        segments in the input data.  This can be quite a large number of pairs so
        we will reduce it by doing a simple preliminary check to eliminate line
        segment pairs that could not possibly cross.  The check is to look at the
        smallest enclosing rectangles (with sides parallel to the axes) for each
        line segment pair and see if they overlap.  If they do then we have to
        compute t1 and t2 (via the A\B computation) to see if the line segments
        cross, but if they don't then the line segments cannot cross.  In a
        typical application, this technique will eliminate most of the potential
        line segment pairs.


        Examples
        --------
        >>> a, b = 1., 2.
        >>> phi  = np.linspace(3, 10, 100)
        >>> x1   = a*phi - b*np.sin(phi)
        >>> y1   = a - b*np.cos(phi)
        >>> x2   = phi
        >>> y2   = np.sin(phi) + 2.
        >>> x,y  = intersection(x1, y1, x2, y2)
        >>> print('{:.1f},{:.1f}'.format(x[0], y[0]))
        6.1,1.8
        >>> print('{:.1f},{:.1f}'.format(x[1], y[1]))
        8.4,2.9


        History
        -------
        Written,  Sukhbinder Singh, Jun 2017 - inspired by revision 1 of Matlab code of Douglas M. Schwarz
                  http://uk.mathworks.com/matlabcentral/fileexchange/11837-fast-and-robust-curve-intersections
        Modified, MC, Nov 2018 - ported to JAMS
    """
    ii,jj = _rectangle_intersection_(x1,y1,x2,y2)
    n = len(ii)

    dxy1 = np.diff(np.c_[x1,y1],axis=0)
    dxy2 = np.diff(np.c_[x2,y2],axis=0)

    T  = np.zeros((4,n))
    AA = np.zeros((4,4,n))
    AA[0:2,2,:]  = -1.
    AA[2:4,3,:]  = -1.
    AA[0::2,0,:] = dxy1[ii,:].T
    AA[1::2,1,:] = dxy2[jj,:].T

    BB = np.zeros((4,n))
    BB[0,:] = -x1[ii].ravel()
    BB[1,:] = -x2[jj].ravel()
    BB[2,:] = -y1[ii].ravel()
    BB[3,:] = -y2[jj].ravel()

    for i in range(n):
        try:
            T[:,i] = np.linalg.solve(AA[:,:,i], BB[:,i])
        except:
            T[:,i] = np.NaN

    in_range = (T[0,:]>=0.) & (T[1,:]>=0.) & (T[0,:]<=1.) & (T[1,:]<=1.)

    xy0 = T[2:,in_range]
    xy0 = xy0.T

    return xy0[:,0], xy0[:,1]


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # # a piece of a prolate cycloid, and am going to find
    # a, b = 1, 2
    # phi = np.linspace(3, 10, 100)
    # x1 = a*phi - b*np.sin(phi)
    # y1 = a - b*np.cos(phi)

    # x2=phi
    # y2=np.sin(phi)+2
    # x,y=intersection(x1,y1,x2,y2)
    # print(x,y)
    # import matplotlib.pyplot as plt
    # plt.plot(x1,y1,c='r')
    # plt.plot(x2,y2,c='g')
    # plt.plot(x,y,'*k')
    # plt.show()



#!/usr/bin/env python
"""
Several forms of the logistic function and their first and second derivatives

The current functions are the following. Each function has a second form with
`_p` appended to the name, where parameters are given as an iterable, e.g.
`logistic` and `logistic_p` = `logistic(x, *p)`:

.. list-table::
   :widths: 15 50
   :header-rows: 1

   * - Function
     - Description
   * - logistic
     - Logistic function `L/(1+exp(-k(x-x0)))`
   * - dlogistic
     - First derivative of logistic function
   * - d2logistic
     - Second derivative of logistic function
   * - logistic_offset
     - logistic function with offset `L/(1+exp(-k(x-x0))) + a`
   * - dlogistic_offset
     - First derivative of logistic function with offset
   * - d2logistic_offset
     - Second derivative of logistic function with offset
   * - logistic2_offset
     - Double logistic function with offset
       `L1/(1+exp(-k1(x-x01))) - L2/(1+exp(-k2(x-x02))) + a`
   * - dlogistic2_offset
     - First derivative of double logistic function with offset
   * - d2logistic2_offset
     - Second derivative of double logistic function with offset

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2015-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

Functions:

.. autosummary::
   logistic
   logistic_p
   dlogistic
   dlogistic_p
   d2logistic
   d2logistic_p
   logistic_offset
   logistic_offset_p
   dlogistic_offset
   dlogistic_offset_p
   d2logistic_offset
   d2logistic_offset_p
   logistic2_offset
   logistic2_offset_p
   dlogistic2_offset
   dlogistic2_offset_p
   d2logistic2_offset
   d2logistic2_offset_p

History
    * Written Mar 2015 by Matthias Cuntz (mc (at) macu (dot) de)
    * Added functions logistic_p and logistic_offset_p,
      Dec 2017, Matthias Cuntz
    * Changed to Sphinx docstring and numpydoc, Dec 2019, Matthias Cuntz
    * Distinguish iterable and array_like parameter types,
      Jan 2020, Matthias Cuntz
    * Make systematically function_p versions of all logistic functions and its
      derivatives, Feb 2020, Matthias Cuntz
    * Split logistic and curvature into separate files,
      May 2020, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""
import numpy as np
import scipy.special as sp


__all__ = ['logistic', 'logistic_p',
           'dlogistic', 'dlogistic_p',
           'd2logistic', 'd2logistic_p',
           'logistic_offset', 'logistic_offset_p',
           'dlogistic_offset', 'dlogistic_offset_p',
           'd2logistic_offset', 'd2logistic_offset_p',
           'logistic2_offset', 'logistic2_offset_p',
           'dlogistic2_offset', 'dlogistic2_offset_p',
           'd2logistic2_offset', 'd2logistic2_offset_p']


# -----------------------------------------------------------
# a/(1+exp(-b(x-c))) - logistic function
def logistic(x, L, k, x0):
    """
    Logistic function

    .. math::
       L/(1+exp(-k(x-x0)))

    Parameters
    ----------
    x : array_like
        Independent variable to evalute logistic function
    L : float
        Maximum of logistic function
    k : float
        Steepness of logistic function
    x0 : float
        Inflection point of logistic function

    Returns
    -------
    float or ndarray
        Logistic function at *x* with maximum *L*, steepness *k* and inflection
        point *x0*

    """
    return L * sp.expit(k * (x - x0))


def logistic_p(x, p):
    """
    Wrapper function for :func:`logistic`: `logistic(x, *p)`

    """
    return logistic(x, *p)


# -----------------------------------------------------------
# 1st derivative of logistic functions
def dlogistic(x, L, k, x0):
    """
    First derivative of :func:`logistic` function

    .. math::
       L/(1+exp(-k(x-x0)))

    which is

    .. math::
       k.L/(2(cosh(k(x-x0))+1))

    Parameters
    ----------
    x : array_like
        Independent variable to evalute derivative of logistic function
    L : float
        Maximum of logistic function
    k : float
        Steepness of logistic function
    x0 : float
        Inflection point of logistic function

    Returns
    -------
    float or ndarray
        First derivative of logistic function at *x* with maximum *L*,
        steepness *k* and inflection point *x0*

    """
    return k * L / (2. * (np.cosh(k * (x - x0)) + 1.))


def dlogistic_p(x, p):
    """
    Wrapper function for :func:`dlogistic`: `dlogistic(x, *p)`

    """
    return dlogistic(x, *p)


# -----------------------------------------------------------
# 2nd derivative of logistic functions
def d2logistic(x, L, k, x0):
    """
    Second derivative of :func:`logistic` function

    .. math::
       L/(1+exp(-k(x-x0)))

    which is

    .. math::
       -k^2.L.sinh(k(x-x0))/(2(cosh(k(x-x0))+1)^2)

    Parameters
    ----------
    x : array_like
        Independent variable to evalute derivative of logistic function
    L : float
        Maximum of logistic function
    k : float
        Steepness of logistic function
    x0 : float
        Inflection point of logistic function

    Returns
    -------
    float or ndarray
        Second derivative of logistic function at *x* with maximum *L*,
        steepness *k* and inflection point *x0*

    """
    return ( -k**2 * L * np.sinh(k * (x - x0)) /
             (2. * (np.cosh(k * (x - x0)) + 1.)**2) )


def d2logistic_p(x, p):
    """
    Wrapper function for :func:`d2logistic`: `d2logistic(x, *p)`

    """
    return d2logistic(x, *p)


# -----------------------------------------------------------
# L/(1+exp(-k(x-x0))) + a - logistic function with offset
def logistic_offset(x, L, k, x0, a):
    """
    Logistic function with offset

    .. math::
       L/(1+exp(-k(x-x0))) + a

    Parameters
    ----------
    x : array_like
        Independent variable to evalute logistic function
    L : float
        Maximum of logistic function
    k : float
        Steepness of logistic function
    x0 : float
        Inflection point of logistic function
    a : float
        Offset of logistic function

    Returns
    -------
    float or ndarray
        Logistic function at *x* with maximum *L*, steepness *k*, inflection
        point *x0* and offset *a*

    """
    return L * sp.expit(k * (x - x0)) + a


def logistic_offset_p(x, p):
    """
    Wrapper function for :func:`logistic_offset`: `logistic_offset(x, *p)`

    """
    return logistic_offset(x, *p)


# -----------------------------------------------------------
# 1st derivative of logistic functions with offset
def dlogistic_offset(x, L, k, x0, a):
    """
    First derivative of :func:`logistic_offset` function

    .. math::
       L/(1+exp(-k(x-x0))) + a

    which is

    .. math::
       k.L/(2(cosh(k(x-x0))+1))

    Parameters
    ----------
    x : array_like
        Independent variable to evalute derivative of logistic function
    L : float
        Maximum of logistic function
    k : float
        Steepness of logistic function
    x0 : float
        Inflection point of logistic function
    a : float
        Offset of logistic function

    Returns
    -------
    float or ndarray
        First derivative of logistic function with offset at *x* with maximum
        *L*, steepness *k*, inflection point *x0*, and offset *a*

    """
    return k * L / (2. * (np.cosh(k * (x - x0)) + 1.))


def dlogistic_offset_p(x, p):
    """
    Wrapper function for :func:`dlogistic_offset`: `dlogistic_offset(x, *p)`

    """
    return dlogistic_offset(x, *p)


# -----------------------------------------------------------
# 2nd derivative of logistic functions with offset
def d2logistic_offset(x, L, k, x0, a):
    """
    Second derivative of :func:`logistic_offset` function

    .. math::
       L/(1+exp(-k(x-x0))) + a

    which is

    .. math::
       -k^2.L.sinh(k(x-x0))/(2(cosh(k(x-x0))+1)^2)

    Parameters
    ----------
    x : array_like
        Independent variable to evalute derivative of logistic function
    L : float
        Maximum of logistic function
    k : float
        Steepness of logistic function
    x0 : float
        Inflection point of logistic function
    a : float
        Offset of logistic function

    Returns
    -------
    float or ndarray
        Second derivative of logistic function at *x* with maximum *L*,
        steepness *k*, inflection point *x0*, and offset *a*

    """
    return ( -k**2 * L * np.sinh(k * (x - x0)) /
             (2. * (np.cosh(k * (x - x0)) + 1.)**2) )


def d2logistic_offset_p(x, p):
    """
    Wrapper function for :func:`d2logistic_offset`: `d2logistic_offset(x, *p)`

    """
    return d2logistic_offset(x, *p)


# -----------------------------------------------------------
# L/(1+exp(-k(x-x0))) + a - logistic function with offset
def logistic2_offset(x, L1, k1, x01, L2, k2, x02, a):
    """
    Double logistic function with offset

    .. math::
       L1/(1+exp(-k1(x-x01))) - L2/(1+exp(-k2(x-x02))) + a

    Parameters
    ----------
    x : array_like
        Independent variable to evalute logistic function
    L1 : float
        Maximum of first logistic function
    k1 : float
        Steepness of first logistic function
    x01 : float
        Inflection point of first logistic function
    L2 : float
        Maximum of second logistic function
    k2 : float
        Steepness of second logistic function
    x02 : float
        Inflection point of second logistic function
    a : float
        Offset of double logistic function

    Returns
    -------
    float or ndarray
        Double Logistic function at *x*

    """
    return L1 * sp.expit(k1 * (x - x01)) - L2 * sp.expit(k2 * (x - x02)) + a


def logistic2_offset_p(x, p):
    """
    Wrapper function for :func:`logistic2_offset`: `logistic2_offset(x, *p)`

    """
    return logistic2_offset(x, *p)


# -----------------------------------------------------------
# 1st derivative of logistic functions with offset
def dlogistic2_offset(x, L1, k1, x01, L2, k2, x02, a):
    """
    First derivative of :func:`logistic2_offset` function

    .. math::
       L1/(1+exp(-k1(x-x01))) - L2/(1+exp(-k2(x-x02))) + a

    which is

    .. math::
       k1.L1/(2(cosh(k1(x-x01))+1)) - k2.L2/(2(cosh(k2(x-x02))+1))

    Parameters
    ----------
    x : array_like
        Independent variable to evalute logistic function
    L1 : float
        Maximum of first logistic function
    k1 : float
        Steepness of first logistic function
    x01 : float
        Inflection point of first logistic function
    L2 : float
        Maximum of second logistic function
    k2 : float
        Steepness of second logistic function
    x02 : float
        Inflection point of second logistic function
    a : float
        Offset of double logistic function

    Returns
    -------
    float or ndarray
        First derivative of double logistic function with offset at *x*

    """
    return ( k1 * L1 / (2. * (np.cosh(k1 * (x - x01)) + 1.)) -
             k2 * L2 / (2. * (np.cosh(k2 * (x - x02)) + 1.)) )


def dlogistic2_offset_p(x, p):
    """
    Wrapper function for :func:`dlogistic2_offset`: `dlogistic2_offset(x, *p)`

    """
    return dlogistic2_offset(x, *p)


# -----------------------------------------------------------
# 2nd derivative of logistic functions with offset
def d2logistic2_offset(x, L1, k1, x01, L2, k2, x02, a):
    """
    Second derivative of :func:`logistic_offset` function

    .. math::
       L1/(1+exp(-k1(x-x01))) - L2/(1+exp(-k2(x-x02))) + a

    which is

    .. math::
       - k1^2.L1.sinh(k1(x-x01)) / (2(cosh(k1(x-x01))+1)^2)
       + k2^2.L2.sinh(k2(x-x02)) / (2(cosh(k2(x-x02))+1)^2)

    Parameters
    ----------
    x : array_like
        Independent variable to evalute logistic function
    L1 : float
        Maximum of first logistic function
    k1 : float
        Steepness of first logistic function
    x01 : float
        Inflection point of first logistic function
    L2 : float
        Maximum of second logistic function
    k2 : float
        Steepness of second logistic function
    x02 : float
        Inflection point of second logistic function
    a : float
        Offset of double logistic function

    Returns
    -------
    float or ndarray
        Second derivative of double logistic function with offset at *x*

    """
    return ( -k1**2 * L1 * np.sinh(k1 * (x - x01)) /
             (2. * (np.cosh(k1 * (x - x01)) + 1.)**2) +
             k2**2 * L2 * np.sinh(k2 * (x - x02)) /
             (2. * (np.cosh(k2 * (x - x02)) + 1.)**2) )


def d2logistic2_offset_p(x, p):
    """
    Wrapper function for
    :func:`d2logistic2_offset`: `d2logistic2_offset(x, *p)`

    """
    return d2logistic2_offset(x, *p)


# -----------------------------------------------------------

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # logistic(1., 1., 0., 2.)
    # # 0.5
    # logistic(1., 1., 2., 1.)
    # # 0.5
    # logistic(2., 1., 1., 1.)
    # # 1./(1.+np.exp(-1.))
    # # 0.7310585786300049
    # logistic_p(1., [1., 0., 2.])
    # logistic_p(1., [1., 2., 1.])
    # logistic_p(2., [1., 1., 1.])
    # logistic_offset(1., 1., 0., 2., 1.)
    # logistic_offset(1., 1., 2., 1., 1.)
    # # 1.5
    # logistic_offset(2., 1., 1., 1., 1.)
    # # 1./(1.+np.exp(-1.)) + 1.
    # # 1.7310585786300049
    # logistic_offset_p(1., [1., 0., 2., 1.])
    # logistic_offset_p(1., [1., 2., 1., 1.])
    # logistic_offset_p(2., [1., 1., 1., 1.])
    # logistic2_offset(1.,  1., 2., 1.,  2., 2., 1.,  1.)
    # # 0.5
    # logistic2_offset_p(1.,  [1., 2., 1.,  2., 2., 1.,  1.])
    # dlogistic(1., 1., 2., 1.)
    # # 0.5
    # dlogistic_offset(1., 1., 2., 1., 1.)
    # # 0.5
    # dlogistic2_offset(1.,  1., 2., 1.,  2., 2., 1.,  1.)
    # # -0.5
    # print(np.around(d2logistic(1., 1., 2., 2.),4))
    # # 0.3199
    # print(np.around(d2logistic_offset(1., 1., 2., 2., 1.),4))
    # # 0.3199
    # print(np.around(d2logistic2_offset(1., 1., 2., 2.,  2., 2., 2.,  1.),4))
    # # -0.3199



#!/usr/bin/env python
"""
This is the unittest for npyio module.

python -m unittest -v tests/test_npyio.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_npyio.py

"""
import unittest


class TestNpyio(unittest.TestCase):
    """
    Tests for npyio.py
    """

    def test_updatez(self):
        import os
        from io import BytesIO
        from tempfile import mkstemp
        import numpy as np
        import pyjams as pj
        from numpy.ma.testutils import assert_equal
        from numpy.testing import temppath

        x = np.zeros((2, 2))
        y = np.zeros((3, 3))
        xnew = np.zeros((4, 4))
        updatez_list = [pj.updatez, pj.updatez_compressed]

        # filename

        # append array
        for updatez in updatez_list:
            with temppath(prefix="np_test_updatez_", suffix=".npz") as tmp:
                np.savez(tmp, x=x)
                ll = np.load(tmp)
                assert 'x' in ll.keys()
                assert_equal(x, ll['x'])
                updatez(tmp, y=y)
                ll = np.load(tmp)
                assert 'x' in ll.keys()
                assert 'y' in ll.keys()
                assert_equal(x, ll['x'])
                assert_equal(y, ll['y'])

        # update array and append array
        for updatez in updatez_list:
            with temppath(prefix="np_test_updatez_", suffix=".npz") as tmp:
                np.savez(tmp, x=x)
                ll = np.load(tmp)
                assert 'x' in ll.keys()
                assert_equal(x, ll['x'])
                updatez(tmp, x=xnew, y=y)
                ll = np.load(tmp)
                assert 'x' in ll.keys()
                assert 'y' in ll.keys()
                assert_equal(xnew, ll['x'])
                assert_equal(y, ll['y'])

        # file is not zipfile
        for updatez in updatez_list:
            with temppath(prefix="np_test_updatez_", suffix=".npz") as tmp:
                updatez(tmp, x=x)
                ll = np.load(tmp)
                assert 'x' in ll.keys()
                assert_equal(x, ll['x'])
                updatez(tmp, x=xnew, y=y)
                ll = np.load(tmp)
                assert 'x' in ll.keys()
                assert 'y' in ll.keys()
                assert_equal(xnew, ll['x'])
                assert_equal(y, ll['y'])

        # array passed as arg
        for updatez in updatez_list:
            with temppath(prefix="np_test_updatez_", suffix=".npz") as tmp:
                updatez(tmp, x=x)
                ll = np.load(tmp)
                assert 'x' in ll.keys()
                assert_equal(x, ll['x'])
                updatez(tmp, xnew, x=xnew, y=y)
                ll = np.load(tmp)
                assert 'x' in ll.keys()
                assert 'y' in ll.keys()
                assert 'arr_0' in ll.keys()
                assert_equal(xnew, ll['x'])
                assert_equal(y, ll['y'])
                assert_equal(xnew, ll['arr_0'])

        # file does not exist yet
        for updatez in updatez_list:
            fd, tmp = mkstemp('.npz')
            os.close(fd)
            os.remove(tmp)
            updatez(tmp, x=x)
            ll = np.load(tmp)
            assert 'x' in ll.keys()
            assert_equal(x, ll['x'])
            updatez(tmp, x=xnew, y=y)
            ll = np.load(tmp)
            assert 'x' in ll.keys()
            assert 'y' in ll.keys()
            assert_equal(xnew, ll['x'])
            assert_equal(y, ll['y'])
            os.remove(tmp)

        # file without .npz suffix
        for updatez in updatez_list:
            fd, tmp = mkstemp()
            os.close(fd)
            os.remove(tmp)
            updatez(tmp, x=x)
            tmpz = tmp + '.npz'
            ll = np.load(tmpz)
            assert 'x' in ll.keys()
            assert_equal(x, ll['x'])
            updatez(tmpz, x=xnew, y=y)
            ll = np.load(tmpz)
            assert 'x' in ll.keys()
            assert 'y' in ll.keys()
            assert_equal(xnew, ll['x'])
            assert_equal(y, ll['y'])
            os.remove(tmpz)

        # Errors

        # path-like objects
        for updatez in updatez_list:
            tmp = BytesIO()
            self.assertRaises(ValueError, updatez, tmp, x=x)

        # un-named variable and keyword have name 'arr_0'
        for updatez in updatez_list:
            with temppath(prefix="np_test_updatez_", suffix=".npz") as tmp:
                # savez raises error
                self.assertRaises(ValueError, updatez, tmp, x, arr_0=x)
                updatez(tmp, x)
                ll = np.load(tmp)
                assert 'arr_0' in ll.keys()
                assert_equal(x, ll['arr_0'])
                # updatez raises error
                self.assertRaises(ValueError, updatez, tmp, x, arr_0=x)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def correlate(x, y, axis=None):
    """
        Computes the cross-correlation function of two series x and y.

        Note that the computations are performed on anomalies (deviations from average).

        Returns the values of the cross-correlation at different lags.
        Lags are given as [0,1,2,...,n,n-1,n-2,...,-2,-1].


        Definition
        ----------
        def correlate(x, y, axis=None):


        Input
        -----
        x          1D or 2D MaskedArray, time series
        y          1D or 2D MaskedArray, time series


        Optional Input
        --------------
        axis       integer (default: None) Axis along which to compute (0 for rows, 1 for cols)
                   If None, the array is flattened first.


        Output
        ------
        Returns the values of the cross-correlation at different lags.
        Lags are given as [0,1,2,...,n,n-1,n-2,...,-2,-1],
        so that the output has a kind of u-shape.


        References
        ----------
        This routine is from Pierre GM on
            http://sourceforge.net/p/matplotlib/mailman/matplotlib-users/?viewmonth=200611&viewday=14
        or
            http://www.mail-archive.com/matplotlib-users@lists.sourceforge.net/msg01487.html



        Examples
        --------
        >>> x = np.sin(np.random.random(10000)) # some data
        >>> y = np.roll(x, 100)                 # shift by 100
        >>> print(np.argmax(correlate(x,y)))          # should be 100
        100

        >>> x = np.sin(np.random.random(2000)).reshape(10,200)
        >>> y = np.roll(x, 100, axis=1)
        >>> ii = np.argmax(correlate(x,y,axis=1),axis=1) # should be 100 and 300
        >>> ii = np.where(ii > 200, ii-200, ii)
        >>> print(ii)
        [100 100 100 100 100 100 100 100 100 100]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Apr 2014 - from Pierre GM at matplotlib-users@lists.sourceforge.net
    """
    assert x.shape == y.shape, "Inconsistent shape !"
    if axis is None:
        if x.ndim > 1:
            x = x.ravel()
            y = y.ravel()
        npad = x.size + y.size
        xanom = (x - x.mean(axis=None))
        yanom = (y - y.mean(axis=None))
        Fx = np.fft.fft(xanom, npad, )
        Fy = np.fft.fft(yanom, npad, )
        iFxy = np.fft.ifft(Fx.conj()*Fy).real
        varxy = np.sqrt(np.inner(xanom,xanom) * np.inner(yanom,yanom))
    else:
        npad = x.shape[axis] + y.shape[axis]
        if axis == 1:
            if x.shape[0] != y.shape[0]:
                raise ValueError("Arrays should have the same length!")
            xanom = (x - x.mean(axis=1)[:,None])
            yanom = (y - y.mean(axis=1)[:,None])
            varxy = np.sqrt((xanom*xanom).sum(1) * (yanom*yanom).sum(1))[:,None]
        else:
            if x.shape[1] != y.shape[1]:
                raise ValueError("Arrays should have the same width!")
            xanom = (x - x.mean(axis=0))
            yanom = (y - y.mean(axis=0))
            varxy = np.sqrt((xanom*xanom).sum(0) * (yanom*yanom).sum(0))
        Fx = np.fft.fft(xanom, npad, axis=axis)
        Fy = np.fft.fft(yanom, npad, axis=axis)
        iFxy = np.fft.ifft(Fx.conj()*Fy,n=npad,axis=axis).real
    #
    return iFxy/varxy


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # x = np.sin(np.random.random(10000)) # some data
    # y = np.roll(x, 100)                 # shift by 100
    # print(np.argmax(correlate(x,y)))          # should be 100

    # x = np.sin(np.random.random(2000)).reshape(10,200)
    # y = np.roll(x, 100, axis=1)
    # ii = np.argmax(correlate(x,y,axis=1),axis=1)   # should be 100 and 300
    # ii = np.where(ii > 200, ii-200, ii)
    # print(ii)



#!/usr/bin/env python
"""
Common test functions for parameter estimation and optimisation algorithms.

Current functions are:
    ackley - >=2 params
        Ackley function, global optimum: 0.0 at origin

    goldstein_price - 2 params
        Goldstein-Price function, global optimum: 3.0 at (0.0,-1.0)

    griewank - 2 or 10 params
        Griewank function, global optimum: 0 at origin

    rastrigin - 2 params
        Rastrigin function, global optimum: -2 at (0,0)

    rosenbrock - 2 params
        Rosenbrock function, global optimum: 0 at (1,1)

    six_hump_camelback - 2 params
       Six-hump Camelback function

       True Optima: -1.031628453489877 at
       (-0.08983, 0.7126) and (0.08983, -0.7126)

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2013-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

Functions:

.. autosummary::
   ackley
   griewank
   goldstein_price
   rastrigin
   rosenbrock
   six_hump_camelback

History
    * Written Oct 2013 by Matthias Cuntz (mc (at) macu (dot) de)
    * Rearrange function library, Mar 2015, Matthias Cuntz
    * Changed to Sphinx docstring and numpydoc, May 2020, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""
import numpy as np


__all__ = ['ackley', 'griewank', 'goldstein_price', 'rastrigin',
           'rosenbrock', 'six_hump_camelback']


# -----------------------------------------------------------

def ackley(x):
    """
    Ackley function (>= 2-D).

    Global Optimum: 0.0 at origin.

    Parameters
    ----------
    x : array
        multi-dimensional x-values (len(x) >= 2)

    Returns
    -------
    float
       Value of Ackley function.

    """
    a = 20.0
    b = 0.2
    c = 2.0 * np.pi

    n  = np.size(x)
    s1 = np.sum(x**2)
    s2 = np.sum(np.cos(c * x))
    f  = (-a * np.exp(-b * np.sqrt(1.0 / n * s1)) - np.exp(1.0 / n * s2) + a +
          np.exp(1.0))

    return f


# -----------------------------------------------------------

def griewank(x):
    """
    Griewank function (2-D or 10-D).

    Global Optimum: 0.0 at origin.

    Parameters
    ----------
    x : array
        multi-dimensional x-values.
        `len(x)=2` or `len(x)=10`.
        `x[i]` bound to [-600, 600] for all i.

    Returns
    -------
    float
       Value of Griewank function.

    """
    nopt = np.size(x)
    # if (nopt == 2) | (nopt == 10):
    xx = x
    if nopt == 2:
        d = 200.0
    else:
        d = 4000.0

    u1 = 0.0
    u2 = 1.0
    for j in range(nopt):
        u1 = u1 + xx[j]**2 / d
        u2 = u2 * np.cos(xx[j] / np.sqrt(float(j + 1)))

    f = u1 - u2 + 1
    return f


# -----------------------------------------------------------

def goldstein_price(x):
    """
    Goldstein-Price function (2-D).

    Global Optimum: 3.0 at (0.0, -1.0).

    Parameters
    ----------
    x : array
        2 x-values. `len(x)=2`.
        `x[i]` bound to [-2, 2] for i=1 and 2.

    Returns
    -------
    float
       Value of Goldstein-Price function.

    """
    x1 = x[0]
    x2 = x[1]
    u1 = (x1 + x2 + 1.0)**2
    u2 = 19. - 14. * x1 + 3. * x1**2 - 14. * x2 + 6. * x1 * x2 + 3. * x2**2
    u3 = (2. * x1 - 3. * x2)**2
    u4 = 18. - 32. * x1 + 12. * x1**2 + 48. * x2 - 36. * x1 * x2 + 27. * x2**2
    u5 = u1 * u2
    u6 = u3 * u4
    f = (1. + u5) * (30. + u6)
    return f


# -----------------------------------------------------------

def rastrigin(x):
    """
    Rastrigin function (2-D).

    Global Optimum: -2.0 at origin.

    Parameters
    ----------
    x : array
        2 x-values. `len(x)=2`.
        `x[i]` bound to [-1, 1] for i=1 and 2.

    Returns
    -------
    float
       Value of Rastrigin function.

    """
    x1 = x[0]
    x2 = x[1]
    f = x1**2 + x2**2 - np.cos(18.0 * x1) - np.cos(18.0 * x2)
    return f


# -----------------------------------------------------------

def rosenbrock(x):
    """
    Rosenbrock function (2-D).

    Global Optimum: 0.0 at (1.0, 1.0).

    Parameters
    ----------
    x : array
        2 x-values. `len(x)=2`.
        `x[1]` bound to [-5, 5].
        `x[2]` bound to [-2, 8].

    Returns
    -------
    float
       Value of Rosenbrock function.

    """

    x1 = x[0]
    x2 = x[1]
    a = 100.0
    f = a * (x2 - x1**2)**2 + (1 - x1)**2
    return f


# -----------------------------------------------------------

def six_hump_camelback(x):
    """
    Six-hump Camelback function (2-D).

    Global Optima: -1.031628453489877
    at (-0.08983, 0.7126) and (0.08983, -0.7126).

    Parameters
    ----------
    x : array
        2 x-values. `len(x)=2`.
        `x[i]` bound to [-5, 5] for i=1 and 2.

    Returns
    -------
    float
       Value of Six-hump Camelback function.

    """
    x1 = x[0]
    x2 = x[1]
    f = ((4 - 2.1 * x1**2 + x1**4 / 3.) * x1**2 + x1 * x2 +
         (-4 + 4 * x2**2) * x2**2)
    return f


# -----------------------------------------------------------

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np
from scipy.stats import t

"""
    Defines common error measures.

    
    Definition
    ----------
    def bias(y_obs,y_mod):      bias
    def mae(y_obs,y_mod):       mean absolute error
    def mse(y_obs,y_mod):       mean squared error
    def rmse(y_obs,y_mod):      root mean squared error
    def nse(y_obs,y_mod):       Nash-Sutcliffe-Efficiency
    def kge(y_obs,y_mod):       Kling-Gupta-Efficiency
    def pear2(y_obs,y_mod):     Squared Pearson correlation coefficient
    def confint(y_obs, p=0.95): Confidence interval of samples
    

    Input
    -----
    y_obs        np.array(N) or np.ma.array(N)
    y_mod        np.array(N) or np.ma.array(N)


    Output
    ------
    measure      float: error measure of respective error function
                        (see definitions for details)

    Restrictions
    ------------
    Deals with masked and unmasked arrays. When nan is found in an unmasked
    array, it will be masked. All measures are applied only on values where
    both, y_obs and y_mod, have valid entries (not masked and not nan)
    
    
    Examples
    --------
    -> see respective function


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014-2017 Arndt Piayda, Stephan Thober, Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written, AP, Jul 2014
    Modified MC, Dec 2014 - use simple formulas that work with normal and masked arrays but do not deal with NaN
    Modified AP, Sep 2015 - add confidence interval
    Modified ST, Nov 2015 - added KGE
    Modified ST, Jan 2017 - added components for KGE
"""

def bias(y_obs,y_mod):
    """
    calculates bias = mean(y_obs) - mean(y_mod)
    
    Examples
    --------
    >>> # Create some data
    >>> y_obs = np.array([12.7867, 13.465, 14.1433, 15.3733, 16.6033])
    >>> y_mod = np.array([12.8087, 13.151, 14.3741, 16.2302, 17.9433])
    >>> # calculate bias
    >>> print(np.round(bias(y_obs, y_mod),2))
    -0.43

    """
    # # check
    # if (y_obs.ndim!=1) or (y_mod.ndim!=1):
    #     raise ValueError('bias: input must be 1D')
    # elif y_obs.size!=y_mod.size:
    #     raise ValueError('bias: input must be of same size')
    # # calc
    # else:
    #     # check if masked or not
    #     try:
    #         temp = y_obs.mask
    #         temp = y_mod.mask
    #     except AttributeError:
    #         y_obs=np.ma.array(y_obs, mask=np.isnan(y_obs))
    #         y_mod=np.ma.array(y_mod, mask=np.isnan(y_mod))
    #     y_modr = np.ma.array(y_mod, mask=y_mod.mask | y_obs.mask)    
    #     y_obsr = np.ma.array(y_obs, mask=y_mod.mask | y_obs.mask)
    # return np.ma.mean(y_obsr) - np.ma.mean(y_modr)
    return y_obs.mean() - y_mod.mean()

def mae(y_obs,y_mod):
    """
    calculates mean absolute error = mean(abs(y_obs - y_mod))
        
    Examples
    --------
    >>> # Create some data
    >>> y_obs = np.array([12.7867, 13.465, 14.1433, 15.3733, 16.6033])
    >>> y_mod = np.array([12.8087, 13.151, 14.3741, 16.2302, 17.9433])
    >>> # calculate mean absolute error
    >>> print(np.round(mae(y_obs, y_mod),2))
    0.55

    """
    # check
    if (y_obs.ndim!=1) or (y_mod.ndim!=1):
        raise ValueError('mae: input must be 1D')
    elif y_obs.size!=y_mod.size:
        raise ValueError('mae: input must be of same size')
    # calc
    else:
        # check if masked or not
        try:
            temp = y_obs.mask
            temp = y_mod.mask
        except AttributeError:
            y_obs=np.ma.array(y_obs, mask=np.isnan(y_obs))
            y_mod=np.ma.array(y_mod, mask=np.isnan(y_mod))
        y_modr = np.ma.array(y_mod, mask=y_mod.mask | y_obs.mask)    
        y_obsr = np.ma.array(y_obs, mask=y_mod.mask | y_obs.mask)
        return np.ma.mean(np.ma.abs(y_obsr-y_modr))

def mse(y_obs,y_mod):
    """
    calculates mean squared error = mean((y_obs - y_mod)**2)
        
    Examples
    --------
    >>> # Create some data
    >>> y_obs = np.array([12.7867, 13.465, 14.1433, 15.3733, 16.6033])
    >>> y_mod = np.array([12.8087, 13.151, 14.3741, 16.2302, 17.9433])
    >>> # calculate mean squared error
    >>> print(np.round(mse(y_obs, y_mod),2))
    0.54

    """
    # # check
    # if (y_obs.ndim!=1) or (y_mod.ndim!=1):
    #     raise ValueError('mse: input must be 1D')
    # elif y_obs.size!=y_mod.size:
    #     raise ValueError('mse: input must be of same size')
    # # calc
    # else:
    #     # check if masked or not
    #     try:
    #         temp = y_obs.mask
    #         temp = y_mod.mask
    #     except AttributeError:
    #         y_obs=np.ma.array(y_obs, mask=np.isnan(y_obs))
    #         y_mod=np.ma.array(y_mod, mask=np.isnan(y_mod))
    #     y_modr = np.ma.array(y_mod, mask=y_mod.mask | y_obs.mask)    
    #     y_obsr = np.ma.array(y_obs, mask=y_mod.mask | y_obs.mask)
    #     return np.ma.mean((y_obsr-y_modr)**2)
    return ((y_obs-y_mod)**2).mean()

def rmse(y_obs,y_mod):
    """
    calculates root mean squared error = sqrt(mean((y_obs - y_mod)**2))
        
    Examples
    --------
    >>> # Create some data
    >>> y_obs = np.array([12.7867, 13.465, 14.1433, 15.3733, 16.6033])
    >>> y_mod = np.array([12.8087, 13.151, 14.3741, 16.2302, 17.9433])
    >>> # calculate root mean squared error
    >>> print(np.round(rmse(y_obs, y_mod),2))
    0.73

    """
    # # check
    # if (y_obs.ndim!=1) or (y_mod.ndim!=1):
    #     raise ValueError('rmse: input must be 1D')
    # elif y_obs.size!=y_mod.size:
    #     raise ValueError('rmse: input must be of same size')
    # # calc
    # else:
    #     # check if masked or not
    #     try:
    #         temp = y_obs.mask
    #         temp = y_mod.mask
    #     except AttributeError:
    #         y_obs=np.ma.array(y_obs, mask=np.isnan(y_obs))
    #         y_mod=np.ma.array(y_mod, mask=np.isnan(y_mod))
    #     y_modr = np.ma.array(y_mod, mask=y_mod.mask | y_obs.mask)
    #     y_obsr = np.ma.array(y_obs, mask=y_mod.mask | y_obs.mask)
    #     return np.ma.sqrt(np.ma.mean((y_obsr-y_modr)**2))
    return ((y_obs-y_mod)**2).mean()**0.5

def nse(y_obs,y_mod):
    """
    calculates Nash-Sutcliffe-Efficiency = 1 - (sum((y_obs - y_mod)**2) / sum((y_obs - mean(y_obs))**2))
       
    Examples
    --------
    >>> # Create some data
    >>> y_obs = np.array([12.7867, 13.465, 14.1433, 15.3733, 16.6033])
    >>> y_mod = np.array([12.8087, 13.151, 14.3741, 16.2302, 17.9433])
    >>> # calculate Nash-Sutcliffe-Efficiency
    >>> print(np.round(nse(y_obs, y_mod),2))
    0.71

    """
    # # check
    # if (y_obs.ndim!=1) or (y_mod.ndim!=1):
    #     raise ValueError('r2: input must be 1D')
    # elif y_obs.size!=y_mod.size:
    #     raise ValueError('r2: input must be of same size')
    # # calc
    # else:
    #     # check if masked or not
    #     try:
    #         temp = y_obs.mask
    #         temp = y_mod.mask
    #     except AttributeError:
    #         y_obs=np.ma.array(y_obs, mask=np.isnan(y_obs))
    #         y_mod=np.ma.array(y_mod, mask=np.isnan(y_mod))
    #     y_modr = np.ma.array(y_mod, mask=y_mod.mask | y_obs.mask)    
    #     y_obsr = np.ma.array(y_obs, mask=y_mod.mask | y_obs.mask)
    #     a = np.ma.sum((y_obsr - y_modr)**2)
    #     b = np.ma.sum((y_obsr - np.ma.mean(y_obsr))**2)
    # return 1. - (a / b)
    return 1. - ((y_obs-y_mod)**2).sum()/((y_obs-y_obs.mean())**2).sum()


def kge(y_obs,y_mod,components=False):
    """
    calculates Kling-Gupta-Efficiency = 1 - sqrt((1-r)**2 + (1-a)**2 + (1-b)**2),
    where r is the Pearson correlation of y_obs and y_mod,
          a is mean(y_mod) / mean(y_obs), and
          b is std(y_mod) / std(y_obs)
    if components is True, then r, a, and b are (in this order) additionally returned to the KGE
       
    Examples
    --------
    >>> # Create some data
    >>> y_obs = np.array([12.7867, 13.465, 14.1433, 15.3733, 16.6033])
    >>> y_mod = np.array([12.8087, 13.151, 14.3741, 16.2302, 17.9433])
    >>> # calculate Kling-Gupta-Efficiency
    >>> print(np.round(kge(y_obs, y_mod),2))
    0.58

    """
    # # check
    # if (y_obs.ndim!=1) or (y_mod.ndim!=1):
    #     raise ValueError('r2: input must be 1D')
    # elif y_obs.size!=y_mod.size:
    #     raise ValueError('r2: input must be of same size')
    # # calc
    # else:
    #     # check if masked or not
    #     try:
    #         temp = y_obs.mask
    #         temp = y_mod.mask
    #     except AttributeError:
    #         y_obs=np.ma.array(y_obs, mask=np.isnan(y_obs))
    #         y_mod=np.ma.array(y_mod, mask=np.isnan(y_mod))
    #     y_modr = np.ma.array(y_mod, mask=y_mod.mask | y_obs.mask)    
    #     y_obsr = np.ma.array(y_obs, mask=y_mod.mask | y_obs.mask)
    #     r = np.ma.corrcoef(y_obsr, y_modr)[0, 1]
    #     a = np.ma.mean(y_modr) / np.ma.mean(y_obsr)
    #     b = np.ma.std(y_modr) / np.ma.std(y_obsr)
    # return 1. - np.sqrt((1 - r)**2 + (1 - a)**2 + (1 - b)**2)
    r = np.corrcoef(y_obs, y_mod)[0, 1]
    alpha = np.std(y_mod) / np.std(y_obs)
    beta = np.mean(y_mod) / np.mean(y_obs)
    if components:
        return 1. - np.sqrt((1 - r)**2 + (1 - beta)**2 + (1 - alpha)**2), r, alpha, beta
    else:
        return 1. - np.sqrt((1 - r)**2 + (1 - beta)**2 + (1 - alpha)**2)


def pear2(y_obs,y_mod):
    """
    calculates squared Pearson correlation coeffcient
     
    Examples
    --------
    >>> # Create some data
    >>> y_obs = np.array([12.7867, 13.465, 14.1433, 15.3733, 16.6033])
    >>> y_mod = np.array([12.8087, 13.151, 14.3741, 16.2302, 17.9433])
    >>> # calculate Squared Pearson correlation coefficient
    >>> print(np.round(pear2(y_obs, y_mod),2))
    0.99

    """
    # # check
    # if (y_obs.ndim!=1) or (y_mod.ndim!=1):
    #     raise ValueError('pear2: input must be 1D')
    # elif y_obs.size!=y_mod.size:
    #     raise ValueError('pear2: input must be of same size')
    # # calc
    # else:
    #     # check if masked or not
    #     try:
    #         temp = y_obs.mask
    #         temp = y_mod.mask
    #     except AttributeError:
    #         y_obs=np.ma.array(y_obs, mask=np.isnan(y_obs))
    #         y_mod=np.ma.array(y_mod, mask=np.isnan(y_mod))
    #     y_modr = np.ma.array(y_mod, mask=y_mod.mask | y_obs.mask)    
    #     y_obsr = np.ma.array(y_obs, mask=y_mod.mask | y_obs.mask)    
    #     return np.corrcoef(y_obsr.compressed(), y_modr.compressed())[0,1]**2
    return ((y_obs-y_obs.mean())*(y_mod-y_mod.mean())).mean()/y_obs.std()/y_mod.std()
    
def confint(y_obs, p=0.95):
    """
    calculates confidence interval of the mean of the sample applying a 
    student-t-distribution to a given probability p (default p=0.95)
     
    Examples
    --------
    >>> # Create some data
    >>> y_obs = np.array([12.7867, 13.465, 14.1433, 15.3733, 16.6033])
    >>> # calculate confident interval
    >>> print(np.round(confint(y_obs)))
    [13. 16.]

    """
    s = y_obs.size
    return np.array(t.interval(p, s-1., loc=y_obs.mean(), scale=y_obs.std()/np.sqrt(s)))

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function

def lif(ifile, noblank=False, comment='', skip=0, maxcol=False):
    """
        Count number of lines in a file.

        Blank (only whitespace) and comment lines can be excluded.


        Definition
        ----------
        def lif(ifile, noblank=False, comment='', skip=0):


        Input
        -----
        ifile         source file name


        Optional input parameters
        -------------------------
        comment      line gets excluded if first character of line is
                     in comment sequence.
                     Sequence can be string, list or tuple
        skip         number of lines to skip at the beginning of
                     the file (default 0)


        Options
        -------
        noblank      if True, exclude all lines that consists only of
                     whitespace characters
        maxcol       if True, return also maximum amount of characters in one line


        Output
        ------
        if maxcol:
            number of lines in file, maximum characters in a line
        else:
            number of lines in file


        Restrictions
        ------------
        Only ascii files.


        Examples
        --------
        # Create some data
        >>> filename = 'test.dat'
        >>> file = open(filename,'w')
        >>> file.writelines('First line\\n')
        >>> file.writelines(' \\n')
        >>> file.writelines('# First comment\\n')
        >>> file.writelines('! Second comment\\n')
        >>> file.writelines('Last line\\n')
        >>> file.close()

        # Count lines
        >>> print(lif(filename))
        5
        >>> print(lif(filename,noblank=True))
        4
        >>> print(lif(filename,comment='#'))
        4
        >>> print(lif(filename,comment='#!'))
        3
        >>> print(lif(filename,comment='#S'))
        4
        >>> print(lif(filename,comment=('#','L')))
        3
        >>> print(lif(filename,comment=['#','!']))
        3
        >>> print(lif(filename,comment='#!',noblank=True))
        2
        >>> print(lif(filename,skip=2))
        3
        >>> print(lif(filename,skip=2,maxcol=True))
        (3, 16)

        # Clean up
        >>> import os
        >>> os.remove(filename)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2009-2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Jul 2009
        Modified, MC, Nov 2012 - maxcol
                  MC, Feb 2013 - ported to Python 3
                  MC, Dec 2014 - changed similar to elegant code of David for head.py
                  MC, Nov 2016 - adapted file handling to Python 2 and 3
    """
    import sys
    if sys.version_info > (3,0):
        import io
        f = ifile if isinstance(ifile,io.TextIOWrapper) else open(ifile,"r")
    else:
        f = ifile if isinstance(ifile,file) else open(ifile,"r")

    # consume lines
    for _ in range(skip):
        f.readline()

    # Read lines
    count = 0
    imax = []
    for line in f:
        if not line:
            break
        if noblank and not line.strip():
            continue
        if line[0] in comment:
            continue
        imax.append(len(line.strip("\n"))) # lines include \0 at the end.
        count += 1

    if f is not ifile:
        f.close()

    if maxcol:
        return count, max(imax)
    else:
        return count


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np


def dewpoint(pres, Celsius=False):
    """
        Calculates the dew point [K] from ambient humidity [Pa].


        Definition
        ----------
        def dewpoint(pres):


        Input
        -----
        pres       actual vapour pressure [Pa]


        Optional Input
        --------------
        Celsius    If True, return degree Celsius instead of Kelvin


        Output
        ------
        Dew point in Kelvin [K]


        Restrictions
        ------------
        None.


        References
        ---------
        Uses Bucks original vapour pressure formulation based on Tetens formula
        Buck, A. L., New equations for computing vapour pressure and enhancement factor,
                     J. Appl. Meteorol., 20, 1527-1532, 1981.

        Examples
        --------
        >>> Ta = 20. + 273.15
        >>> from jams import esat
        >>> es = esat(Ta, formula='Buck_original')
        >>> from autostring import astr
        >>> print(astr(dewpoint(es),3,pp=True))
        293.150

        >>> print(astr(dewpoint(es, Celsius=True),3,pp=True))
        20.000


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2013 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Jan 2012
        Modified, MC, Feb 2013 - ported to Python 3
    """
    pw  = 611.21
    c1  = 240.97
    c2  = 17.502
    w   = np.log(pres/pw)
    out = w*c1/(c2-w)
    if Celsius:
        return out
    else:
        import pyjams.const as const
        return out + const.T0


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
Physical, mathematical, computational, isotope, and material constants.

Defines the following constants:
    Mathematical
        Pi, Pi2, Pi3, TwoPi, Sqrt2, pi, pi2, pi3, Twopi

    Physical
        Gravity, T0, P0, T25, sigma, R, R_air, R_H2O, Na, REarth

    Isotope
        R13VPDB, R18VSMOW, R2VSMOW

    Computational
        tiny, huge, eps

    Material
        mmol_co2, mmol_h2o, mmol_air,
        density_quartz, cheat_quartz, cheat_water, cheat_air,
        latentheat_vaporization

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2012-2021 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

Constants:

.. autosummary::
   Pi
   Pi2
   Pi3
   TwoPi
   pi
   pi2
   pi3
   Twopi
   Sqrt2
   sqrt2
   gravity
   T0
   P0
   T25
   sigma
   R
   Rair
   Rh2o
   Na
   kB
   REarth
   mmol_co2
   molmass_co2
   mmol_h2o
   molmass_h2o
   mmol_air
   molmass_air
   density_quartz
   cheat_quartz
   cheat_water
   cheat_air
   latentheat_vaporization
   R13VPDB
   R18VSMOW
   R2VSMOW
   tiny
   huge
   eps

History
    * Written Jan 2012 by Matthias Cuntz (mc (at) macu (dot) de)
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Added dielectric constant for water, Mar 2014, Arndt Piayda
    * Added heat capacities for air, water and quartz as well as density of
      quartz, Sep 2014, Arndt Piayda
    * Added Pi3=pi/3, R13VPDB, R18VSMOW, R2VSMOW, Mar 2015, Matthias Cuntz
    * Renamed heat capacities, molar masses, density of quartz,
      Mar 2015, Matthias Cuntz
    * Moved calculation of dielectric constant of water to own routine,
      Mar 2015, Matthias Cuntz
    * Added computational constants such as tiny=np.finfo(np.float).tiny,
      Nov 2016, Matthias Cuntz
    * Added gas constants for dry air and water, May 2017, RL
    * Using numpy docstring format, May 2020, Matthias Cuntz
    * Added lowercase version of pi constants, May 2020, Matthias Cuntz
    * Updated all constants related to gases for redefinition of
      SI units of 2019, Jan 2023, Matthias Cuntz
    * Renamed some constant for greater consistency, Jan 2023, Matthias Cuntz

"""
import numpy as np


__all__ = [
    'Pi', 'Pi2', 'Pi3', 'TwoPi', 'pi', 'pi2', 'pi3', 'Twopi', 'Sqrt2', 'sqrt2',
    'gravity', 'T0', 'P0', 'T25', 'sigma', 'R', 'Rair', 'Rh2o',
    'Na', 'REarth', 'mmol_co2', 'molmass_co2', 'mmol_h2o', 'molmass_h2o',
    'mmol_air', 'molmass_air',
    'density_quartz', 'cheat_quartz', 'cheat_water', 'cheat_air',
    'latentheat_vaporization',
    'R13VPDB', 'R18VSMOW', 'R2VSMOW',
    'tiny', 'huge', 'eps']


# Mathematical
Pi = 3.141592653589793238462643383279502884197
r"""
Mathematical constant :math:`\pi`
"""
pi = 3.141592653589793238462643383279502884197
r"""
Mathematical constant :math:`\pi`
"""
Pi2 = 1.57079632679489661923132169163975144209858
r"""
Mathematical constant :math:`\pi/2`
"""
pi2   = 1.57079632679489661923132169163975144209858
r"""
Mathematical constant :math:`\pi/2`
"""
Pi3   = 1.0471975511965977461542144610931676280656
r"""
Mathematical constant :math:`\pi/3`
"""
pi3   = 1.0471975511965977461542144610931676280656
r"""
Mathematical constant :math:`\pi/3`
"""
TwoPi = 6.283185307179586476925286766559005768394
r"""
Mathematical constant :math:`2\pi`
"""
Twopi = 6.283185307179586476925286766559005768394
r"""
Mathematical constant :math:`2\pi`
"""
Sqrt2 = 1.41421356237309504880168872420969807856967
r"""
Mathematical constant :math:`\sqrt{2}`
"""
sqrt2 = 1.41421356237309504880168872420969807856967
r"""
Mathematical constant :math:`\sqrt{2}`
"""

# Physical
gravity = 9.80665
r"""
Standard average Earth's gravity (:math:`m^2 s^{-1}`)
"""
T0 = 273.15
"""
0 degree Celsius in Kelvin.
Conversion constant from Celsius to Kelvin.
"""
P0 = 101325.
"""
Standard pressure (Pa)
"""
T25 = 298.15
r"""
Standard ambient temperature of 25 :math:`^\circ C` in Kelvin [K]
"""
sigma = 5.67e-08
"""
Stefan-Boltzmann constant (:math:`W m^{-2} K^{-4}`)
"""
R = 8.31446261815324
r"""
Ideal gas constant R = Na*kB (:math:`J K^{-1} mol^{-1}`)
"""
Rair = 287.05799595894405
r"""
Specific gas constant of dry air Rair = R/molmass_air (:math:`J K^{-1} kg^{-1}`)
"""
Rh2o = 461.52280831345604
r"""
Gas constant of water vapour Rh2o = R/molmass_h2o (:math:`J K^{-1} kg^{-1}`)
"""
Na = 6.02214076e23
r"""
Avogadro number (:math:`mol^{-1}`)
"""
kB = 1.380649e-23
r"""
Boltzmann constant (:math:`J K^{-1}`)
"""
REarth = 6371009.
"""
Radius of the Earth (m)
"""

# Material
mmol_co2 = 44.009
r"""
Molar mass of :math:`CO_2` (:math:`g mol^{-1}`)
"""
molmass_co2 = 44.009e-3
r"""
Molar mass of :math:`CO_2` (:math:`kg mol^{-1}`)
"""
mmol_h2o = 18.01528
r"""
Molar mass of water (:math:`g mol^{-1}`)
"""
molmass_h2o = 18.01528e-3
r"""
Molar mass of water (:math:`kg mol^{-1}`)
"""
mmol_air = 28.9644
r"""
Molar mass of dry air of standard atmosphere (:math:`g mol^{-1}`)
"""
molmass_air = 28.9644e-3
r"""
Molar mass of dry air (:math:`kg mol^{-1}`)
"""
# from Cambell G (1985) Soil Physics with BASIC, Elsevier Science
density_quartz = 2.65
r"""
Density of quartz (:math:`g cm^{-3}`)
"""
cheat_quartz = 800.
r"""
Heat capacity of quartz (:math:`J kg^-1 K^-1`)
"""
cheat_water = 4180.
r"""
Heat capacity of water (:math:`J kg^{-1} K^{-1}`)
"""
cheat_air = 1010.
r"""
Heat capacity of air (:math:`J kg^{-1} K^{-1}`)
"""
latentheat_vaporization = 2.45e6
r"""
Latent heat of vaporization of water (:math:`J kg^{-1}`)
"""

# Isotope
R13VPDB = 0.0112372
r"""
:math:`^{13}C` isotope ratio of VPDB
"""
R18VSMOW = 2005.2e-6
r"""
:math:`^{18}O` isotope ratio of VSMOW
"""
R2VSMOW = 155.76e-6
r"""
Deuterium= :math:`^{2}H` isotope ratio of VSMOW
"""

# Computational
eps = np.finfo(float).eps
r"""
Numerical precision of floats.
The difference between 1.0 and the next smallest representable float larger
than 1.0. For example, for 64-bit binary floats in the IEEE-754 standard,
:math:`eps = 2^{-52}`, approximately 2.22e-16.
"""
huge = np.finfo(float).max
"""
The largest representable floating point number.
"""
tiny = np.finfo(float).tiny
"""
The smallest positive floating point number with full precision.
"""


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

__all__ = ['ellipse_area']

def ellipse_area(a, b=None):
    """
        Area of ellipse with major and minor axes a and b.


        Definition
        ----------
        def ellipse_area(a, b=None):


        Input
        -----
        a          semi-major axis


        Optional Input
        --------------
        b          semi-minor axis


        Output
        ------
        area of ellipse or circle if semi-minor axis not given


        Restrictions
        ------------
        None


        Examples
        --------
        >>> from autostring import astr
        >>> area = ellipse_area(1, 1)
        >>> print(astr(area, 3))
        3.142

        >>> area = ellipse_area(2)
        >>> print(astr(area, 3))
        12.566
        

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Oct 2014 - in Python seminar
    """
    if b is None:
        return np.pi*a*a
    else:
        return np.pi*a*b


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # from autostring import astr
    # area = ellipse_area(1, 1)
    # print(astr(area, 3))

    # area = ellipse_area(2)
    # print(astr(area, 3))



#!/usr/bin/env python
import numpy as np
from pyjams.jams.eddybox import gapfill
from pyjams.air_humidity import esat
from pyjams.jams.date2dec import date2dec
from pyjams.jams.autostring import astr


def fluxfill(fluxfile, metfile, outdir, rg, tair, rh, delimiter=[',',','],
             skiprows=[1,1], format=['ascii','ascii'], undef=-9999, plot=False):
    '''
    Wrapper function for gapfill with file management and plotting.


    Definition
    ----------
    fluxfill(fluxfile, metfile, outdir, rg, tair, rh, delimiter=[',',','],
             skiprows=[1,1], format=['ascii','ascii'], undef=-9999, plot=False):


    Input
    -----
    fluxfile    str, path and file name of fluxflag output file containing
                fluxes and flags
    metfile     str, path and file name of the meteorology file (must be in
                sync with fluxfile)
    outdir      str, path of the output folder
    rg          int, column number of global radiation [W m-2] in metfile,
                column number starts with 0 which is first data column.
    tair        int, column number of air temperature [deg C] in metfile, column
                number starts with 0 which is first data column.
    rh          int, column number of relative humidity [%] in metfile, column
                number starts with 0 which is first data column.


    Optional Input
    --------------
    delimiter   list of str, delimiters of fluxfile and metfile
                (default: [',',','])
    skiprows    list of int, lines to skip at the beginning of fluxfile and
                metfile, e.g. header lines (default: [1,1])
    format      list of str, time formats of fluxfile and metfile, 'ascii' and
                'eng' possible (default: ['ascii','ascii'])
    undef       int/float, missing value of fluxfile and metfile
                (default: -9999, np.nan not possible)
    plot        bool, if True performs plotting (default: False)


    Output
    ------
    fluxfilled.csv file containing fluxes with original flags and quality flags
                   of the gap filling. Fluxes are filled where flag>1.
    fluxfilled.pdf  plot of each gap filled flux.


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Aug 2014
    '''

    ###########################################################################
    # reading input files
    d = np.loadtxt(fluxfile, dtype='|S100', delimiter=delimiter[0], skiprows=skiprows[0])
    m = np.loadtxt(metfile,  dtype='|S100', delimiter=delimiter[1], skiprows=skiprows[1])

    assert (d.shape[1]==11) | (d.shape[1]==19), 'fluxfill: fluxfile must be from fluxflag or profile2storage and have 11 or 19 cols'

    if format[0]=='ascii':
        datev   = date2dec(ascii=d[:,0])
    elif format[0]=='eng':
        datev   = date2dec(eng=d[:,0])
    else:
        raise ValueError('fluxfill: unknown format')
    if format[1]=='ascii':
        datem   = date2dec(ascii=m[:,0])
    elif format[1]=='eng':
        datem   = date2dec(eng=m[:,0])
    else:
        raise ValueError('fluxfill: unknown format')

    val = np.where(d[:,1:]=='', str(undef), d[:,1:]).astype(np.float)
    met = np.where(m[:,1:]=='', str(undef), m[:,1:]).astype(np.float)

    ###########################################################################
    # assign variables
    if (d.shape[1]==11):
        data      = val[:,0::2] # corr. H, corr. LE, corr. E, corr. C, tau
        data_flag = (val[:,1::2]>1) | (val[:,0::2]==undef)
    else:
        data      = val[:,[0,1, 3,4, 6,7, 9,10, 12]] # corr. H, corr. H+s, corr. LE, corr. LE+s, corr. E, corr. E+s, corr. C, corr. C+s, tau
        data_flag = np.repeat((val[:,[2,5,8,11,13]]>1) | (val[:,[0,1, 3,4, 6,7, 9,10, 12]]==undef), 2, axis=1)[:,:-1]
    rg        = met[:,rg]
    rg_flag   = rg==undef
    tair      = met[:,tair]
    tair_flag = tair==undef
    rh        = met[:,rh]
    rh_flag   = rh==undef
    vpd       = np.empty_like(tair)
    vpd[tair_flag | rh_flag]    = undef
    vpd[~(tair_flag | rh_flag)] = (1.-rh[~(tair_flag | rh_flag)]/100.)*esat(tair[~(tair_flag | rh_flag)]+273.14)
    vpd_flag  = vpd==undef

    flux = np.zeros_like(data)
    flag = np.zeros_like(data_flag, dtype=np.int)

    ###########################################################################
    if plot:
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_pdf as pdf
        pp1 = pdf.PdfPages(outdir+'/fluxfilled.pdf')

    ###########################################################################
    # do gapfill
    for i in range(data.shape[1]):
        flux[:,i], flag[:,i] = gapfill.gapfill(datev, data[:,i], rg, tair, vpd,
                                       data_flag[:,i], rg_flag, tair_flag,
                                       vpd_flag, rg_dev=50., tair_dev=2.5,
                                       vpd_dev=5., longgap=60, fullday=False,
                                       undef=undef, ddof=1, err=False,
                                       shape=False)

        #######################################################################
        # plot
        if plot:
            majticks = mpl.dates.MonthLocator(bymonthday=1)
            format_str='%d %m %Y %H:%M'
            date01 = date2dec(yr=1, mo=1, dy=2, hr=0, mi=0, sc=0)

            fig1 = plt.figure(1)
            sub1 = fig1.add_subplot(111)
            l1 =sub1.plot(datev-date01, flux[:,i], '-g')
            l2 =sub1.plot(datev-date01, np.ma.array(data[:,i],mask=data_flag[:,i]), '-b')

            sub1.set_xlim(datev[0]-date01,datev[-1]-date01)
            sub1.xaxis.set_major_locator(majticks)
            sub1.xaxis.set_major_formatter(mpl.dates.DateFormatter(format_str))
            fig1.autofmt_xdate()

            plt.show()
            fig1.savefig(pp1, format='pdf')

    if plot:
        pp1.close()

    ###########################################################################
    # prepare output and save file
    flux_str       = np.array([['%11.5f'%x for x in y] for y in flux])

    if (d.shape[1]==11):
        header         = np.array(['          H', '   Hflag', '     Hgf',
                                   '         LE', '  LEflag', '    LEgf',
                                   '          E', '   Eflag', '     Egf',
                                   '          C', '   Cflag', '     Cgf',
                                   '        TAU', ' TAUflag', '   TAUgf'])
        output         = np.hstack((d[:,0:1], flux_str.repeat(3, axis=1)))
        output[:,2::3] = astr(val[:,1::2].astype(int), prec=8)
        output[:,3::3] = astr(flag, prec=8)
    else:
        header         = np.array(['          H', '       H+sT', '   Hflag', '     Hgf',
                                   '         LE', '     LE+sLE', '  LEflag', '    LEgf',
                                   '          E', '       E+sE', '   Eflag', '     Egf',
                                   '          C', '       C+sC', '   Cflag', '     Cgf',
                                   '        TAU',    ' TAUflag',             '   TAUgf',
                                   '         sT', '        sLE', '         sE', '         sC'])
        output                   = np.hstack((d[:,0:1], np.insert(flux_str, [2,2,4,4,6,6,8,8], flux_str[:,:-1], axis=1),
                                              flux_str[:,-1:].repeat(2, axis=1), d[:,-4:]))
        output[:,[3,7,11,15,18]] = astr(val[:,[2,5,8,11,13]].astype(int), prec=8)
        output[:,[4,8,12,16,19]] = astr(flag[:,[0,2,4,6,8]], prec=8)

    np.savetxt('%s/fluxfilled.csv'%outdir,
               np.vstack((np.concatenate((['            date'], header))[np.newaxis,:],
                          output)), '%s', delimiter=',')

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
import numpy as np


def spike(datin, thresh=7, toler=0, length=15):
    """
        spike filter which looks for spikes which have a first step > thresh and then a 'plateau' for <= length time steps and come back
        to original value within a tolerance of toler. (something like a rectangular shape, but the 'plateau' does not have to be flat,
        it just needs to exceed the threshold without crossing the value before the spike).
        Returns list with indices of detected spikes.


        Definition
        ----------
        def spike(datin, thresh=7, toler=0, length=15):


        Input
        -----
        datin      1d array or list


        Optional Input
        --------------
        thresh     spike threshold
        toler      tolerance value for the abs difference between first value before spike and first value after spike
        length     maximum length of spike plateau


        Output
        ------
        list with indices of all spikes detected


        Restrictions
        ------------
        --


        Examples
        --------
        >>> y = [1, 1, 20,1, 2, 1,15,1, 10, 11, 13, 10, 1.5, 1, 2, 1, 1, -7, -8, 1.4, 3, 1, 1, 1, -50, 1, 1, 1]
        >>> spike(y,thresh=2,toler=0.6,length=10)
        [2, 6, 8, 9, 10, 11, 17, 18, 24]


        >>> y = [0.1, 0.1, 0.2, 0.1, 0.2, 0.1, 1.5, 0.1, 1.0, 1.1, 1.3, 1.0, 0.15, 0.1, 0.2, 0.1, 0.1, -0.7, -0.8, 0.14, 0.3, 0.1, 0.1, 0.1, -5.0, 0.1, 0.1, 0.1]
        >>> spike(y,thresh=0.2,toler=0.06,length=3)
        [6, 17, 18, 24]


        >>> y = [0.1, 0.1, 0.2, 0.1, 0.2, 0.1, 1.5, 0.2, 1.0, 1.1, 1.3, 1.0, 0.15, 0.1, 0.2, 0.1, 0.1, -0.7, 0.8, 0.14, 0.3, 0.1, 0.1, 0.1, -5.0, 0.1, 0.1, 0.1]
        >>> spike(y,thresh=0.2,toler=0.06,length=3)
        [24]

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Benjamin Dechant

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  BD, Sep 2016
                  MC, Nov 2016 - ported to Python 3
                  DS, Aug 2017 - reduced verbosity 

    """
    debug = False
    cmp = lambda a, b: ((a > b) - (a < b)) # redefine cmp to be available in Python 3

    diff = [ datin[k+1]-datin[k] for k in range(len(datin)-1) ]  # differences
    ipos0 = [ k for k,a in enumerate(diff) if abs(a) > thresh ]  # get index position of spikes
    if len(ipos0)>0:
        ipos = ipos0[0]                                          # select first spike
        maxlen = length                                          # set max. length of spike plateau
        spike_pos_all = []
        while ipos < len(datin)-1:
            # if debug: print('ipos=',ipos)
            for i in range(1,maxlen+1):
                # if debug: print('i=',i)
                spike_pos = []
                if (ipos+i+1) > (len(datin)-1):
                    ipos = len(datin)
                    break
                tm = [ abs(datin[ipos]-datin[ipos+v]) for v in  list(range(2,i+2)) ] # diff between first val before spike and all candidates for first val after spike
                tms = [ datin[ipos]-datin[ipos+v] for v in  list(range(2,i+1)) ] # diff between first val before spike and all candidates for first val after spike (with sign)
                if len(tm)==1 and tm[0] < toler:
                    spike_pos = [ipos+1]
                    spike_pos_all.append(spike_pos)
                    # if debug: print('found peak 1 ', spike_pos)
                    break
                elif (len(tm)>1 and
                      all([ tt > thresh for tt in tm[0:-2] ]) and
                      all([ cmp(val,0) == cmp(datin[ipos]-datin[ipos+1],0) for n, val in enumerate(tms) ]) and
                      tm[-1] < toler): # check thresh, tolerance and no switching of sign
                    spike_pos = list(range(ipos+1,ipos+i+1))
                    spike_pos_all.append(spike_pos)
                    # if debug: print('found peak 2',spike_pos)
                    break
            ipos1 = [k for k,a in enumerate(diff) if abs(a) > thresh  and k >= ipos + i +1] # get index position of next spikes
            if len(ipos1)> 0:
                ipos = ipos1[0]
            else:
                break

        # print('no. of spikes: '+str(len(spike_pos_all)))
        spike_pos_all = [item for sublist in spike_pos_all for item in sublist]    # create list without sublists
        # print('no. of data points: '+str(len(spike_pos_all)))
        return spike_pos_all

    else:
        return []



if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



"""
Collection of color palettes and continuous color maps

:copyright: Copyright 2021- Matthias Cuntz, see AUTHORS.md for details.
:license: MIT License, see LICENSE for details.

Subpackages
===========
.. autosummary::
   brewer_palettes
   mathematica_palettes
   ncl_palettes
   oregon_palettes
   pyjams_palettes
   sron2012_palettes
   sron_palettes
   ufz_palettes
   color

"""
# colour palettes
from .brewer_palettes import brewer_sequential, brewer_diverging
from .brewer_palettes import brewer_qualitative
from .mathematica_palettes import mathematica_rainbow
from .ncl_palettes import ncl_large, ncl_small, ncl_meteo_swiss
from .oregon_palettes import oregon_sequential, oregon_diverging
from .oregon_palettes import oregon_qualitative
from .pyjams_palettes import pyjams_cmaps
from .sron2012_palettes import sron2012_colors, sron2012_functions
from .sron_palettes import sron_colors, sron_colormaps, sron_functions
from .ufz_palettes import ufz_colors
# get, show, print colors and color palettes
from .color import get_color, print_colors
from .color import get_cmap, print_palettes, show_palettes


__all__ = ['brewer_sequential', 'brewer_diverging', 'brewer_qualitative',
           'mathematica_rainbow',
           'ncl_large', 'ncl_small', 'ncl_meteo_swiss',
           'oregon_sequential', 'oregon_diverging', 'oregon_qualitative',
           'pyjams_cmaps',
           'sron2012_colors', 'sron2012_functions',
           'sron_colors', 'sron_colormaps', 'sron_functions',
           'ufz_colors',
           'get_color', 'print_colors',
           'get_cmap', 'print_palettes', 'show_palettes',
           ]
